# -*- coding: utf-8 -*-

import itertools
import logging
from functools import cached_property
from pathlib import Path

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from numpy.typing import ArrayLike
from pybader.interface import Bader
from pymatgen.io.vasp import VolumetricData
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pyrho.pgrid import PGrid

# from scipy.ndimage import binary_erosion


class Grid(VolumetricData):
    """
    This class is a wraparound for Pymatgen's VolumetricData class with additional
    properties and methods useful to the badelf algorithm
    """

    @property
    def total(self):
        return self.data["total"]

    @total.setter
    def total(self, new_total):
        self.data["total"] = new_total

    @property
    def diff(self):
        return self.data["diff"]

    @diff.setter
    def diff(self, new_diff):
        self.data["diff"] = new_diff

    @property
    def shape(self):
        return np.array(self.total.shape)

    @property
    def matrix(self):
        """
        A 3x3 matrix defining the a, b, and c sides of the unit cell
        """
        return self.structure.lattice.matrix

    @property
    def a(self):
        """
        The cartesian coordinates for the lattice vector "a"
        """
        return self.matrix[0]

    @property
    def b(self):
        """
        The cartesian coordinates for the lattice vector "b"
        """
        return self.matrix[1]

    @property
    def c(self):
        """
        The cartesian coordinates for the lattice vector "c"
        """
        return self.matrix[2]

    @property
    def frac_coords(self):
        """
        Array of fractional coordinates for each atom.
        """
        return self.structure.frac_coords

    @property
    def all_voxel_frac_coords(self):
        """
        The fractional coordinates for all of the voxels in the grid
        """
        a, b, c = self.shape
        voxel_indices = np.indices(self.shape).reshape(3, -1).T
        frac_coords = voxel_indices.copy().astype(float)
        frac_coords[:, 0] /= a
        frac_coords[:, 1] /= b
        frac_coords[:, 2] /= c
        return frac_coords

    @cached_property
    def voxel_dist_to_origin(self):
        frac_coords = self.all_voxel_frac_coords
        cart_coords = self.get_cart_coords_from_frac_full_array(frac_coords)
        corners = [
            np.array([0, 0, 0]),
            self.a,
            self.b,
            self.c,
            self.a + self.b,
            self.a + self.c,
            self.b + self.c,
            self.a + self.b + self.c,
        ]
        distances = []
        for corner in corners:
            voxel_distances = np.linalg.norm(cart_coords - corner, axis=1).round(6)
            distances.append(voxel_distances)
        min_distances = np.min(np.column_stack(distances), axis=1)
        min_distances = min_distances.reshape(self.shape)
        return min_distances

    @property
    def voxel_volume(self):
        """
        The volume of each voxel in the grid
        """
        volume = self.structure.volume
        voxel_num = np.prod(self.shape)
        return volume / voxel_num

    @property
    def voxel_num(self):
        """
        The number of voxels in the grid
        """
        return self.shape.prod()

    @property
    def max_voxel_dist(self):
        """
        Finds the maximum distance a voxel can be from a dividing plane that still
        allows for the possibility that the voxel is intercepted by the plane.

        Returns:
            The maximum distance a voxel can be from a dividing plane and still
            be intercepted by the plane.
        """
        # We need to find the coordinates that make up a single voxel. This
        # is just the cartesian coordinates of the unit cell divided by
        # its grid size
        end = [0, 0, 0]
        vox_a = [x / self.shape[0] for x in self.a]
        vox_b = [x / self.shape[1] for x in self.b]
        vox_c = [x / self.shape[2] for x in self.c]
        # We want the three other vertices on the other side of the voxel. These
        # can be found by adding the vectors in a cycle (e.g. a+b, b+c, c+a)
        vox_a1 = [x + x1 for x, x1 in zip(vox_a, vox_b)]
        vox_b1 = [x + x1 for x, x1 in zip(vox_b, vox_c)]
        vox_c1 = [x + x1 for x, x1 in zip(vox_c, vox_a)]
        # The final vertex can be found by adding the last unsummed vector to any
        # of these
        end1 = [x + x1 for x, x1 in zip(vox_a1, vox_c)]
        # The center of the voxel sits exactly between the two ends
        center = [(x + x1) / 2 for x, x1 in zip(end, end1)]
        # Shift each point here so that the origin is the center of the
        # voxel.
        voxel_vertices = []
        for vector in [
            center,
            end,
            vox_a,
            vox_b,
            vox_c,
            vox_a1,
            vox_b1,
            vox_c1,
            end,
        ]:
            new_vector = [(x - x1) for x, x1 in zip(vector, center)]
            voxel_vertices.append(new_vector)

        # Now we need to find the maximum distance from the center of the voxel
        # to one of its edges. This should be at one of the vertices.
        # We can't say for sure which one is the largest distance so we find all
        # of their distances and return the maximum
        max_distance = max([np.linalg.norm(vector) for vector in voxel_vertices])
        return max_distance

    @property
    def permutations(self):
        """
        The permutations for translating a voxel coordinate to nearby unit
        cells. This is necessary for the many voxels that will not be directly
        within an atoms partitioning.

        Returns:
            A list of voxel permutations unique to the grid dimensions.
        """
        a, b, c = self.shape
        permutations = [
            (t, u, v)
            for t, u, v in itertools.product([-a, 0, a], [-b, 0, b], [-c, 0, c])
        ]
        # sort permutations. There may be a better way of sorting them. I
        # noticed that generally the correct site was found most commonly
        # for the original site and generally was found at permutations that
        # were either all negative/0 or positive/0
        permutations_sorted = []
        for item in permutations:
            if all(val <= 0 for val in item):
                permutations_sorted.append(item)
            elif all(val >= 0 for val in item):
                permutations_sorted.append(item)
        for item in permutations:
            if item not in permutations_sorted:
                permutations_sorted.append(item)
        permutations_sorted.insert(0, permutations_sorted.pop(7))
        return permutations_sorted

    @property
    def voxel_resolution(self):
        volume = self.structure.volume
        number_of_voxels = self.shape.prod()
        return number_of_voxels / volume

    @property
    def equivalent_atoms(self):
        return SpacegroupAnalyzer(self.structure).get_symmetry_dataset()[
            "equivalent_atoms"
        ]

    def get_2x_supercell(self, data: ArrayLike):
        """
        Duplicates data with the same dimensions as the grid to make a 2x2x2
        supercell

        Args:
            data (ArrayLike):
                The data to duplicate. Must have the same dimensions as the
                grid.

        Returns:
            The duplicated data
        """
        raveled_data = data.ravel()
        voxel_indices = np.indices(data.shape).reshape(3, -1).T
        transformations = [
            [0, 0, 0],  # -
            [1, 0, 0],  # x
            [0, 1, 0],  # y
            [0, 0, 1],  # z
            [1, 1, 0],  # xy
            [1, 0, 1],  # xz
            [0, 1, 1],  # yz
            [1, 1, 1],  # xyz
        ]
        transformations = np.array(transformations)
        transformations = self.get_voxel_coords_from_frac(transformations.T).T
        supercell = np.zeros(np.array(data.shape) * 2)
        for transformation in transformations:
            transformed_indices = (voxel_indices + transformation).astype(int)
            x = transformed_indices[:, 0]
            y = transformed_indices[:, 1]
            z = transformed_indices[:, 2]
            supercell[x, y, z] = raveled_data
        return supercell

    def get_voxels_in_radius(self, radius: float, voxel: ArrayLike):
        """
        Gets the indices of the voxels in a radius around a voxel

        Args:
            radius (float):
                The radius in Angstroms around the voxel

            voxel (ArrayLike):
                The voxel coordinates of the voxel to find the sphere around

        Returns:
            The voxel indices of the voxels within the provided radius
        """
        voxel = np.array(voxel)
        # Get the distance from each voxel to the origin
        voxel_distances = self.voxel_dist_to_origin

        # Get the indices that are within the radius
        sphere_indices = np.where(voxel_distances <= radius)
        sphere_indices = np.column_stack(sphere_indices)

        # Get indices relative to the voxel
        sphere_indices = sphere_indices + voxel
        # adjust voxels to wrap around grid
        # line = [[round(float(a % b), 12) for a, b in zip(position, grid_data.shape)]]
        new_x = (sphere_indices[:, 0] % self.shape[0]).astype(int)
        new_y = (sphere_indices[:, 1] % self.shape[1]).astype(int)
        new_z = (sphere_indices[:, 2] % self.shape[2]).astype(int)
        sphere_indices = np.column_stack([new_x, new_y, new_z])
        # return new_x, new_y, new_z
        return sphere_indices

    def get_voxels_transformations_to_radius(self, radius: float):
        """
        Gets the transformations required to move from a voxel to the voxels
        surrounding it within the provided radius

        Args:
            radius (float):
                The radius in Angstroms around the voxel

        Returns:
            An array of transformations to add to a voxel to get to each of the
            voxels within the radius surrounding it
        """
        # Get voxels around origin
        voxel_distances = self.voxel_dist_to_origin
        # sphere_grid = np.where(voxel_distances <= radius, True, False)
        # eroded_grid = binary_erosion(sphere_grid)
        # shell_indices = np.where(sphere_grid!=eroded_grid)
        shell_indices = np.where(voxel_distances <= radius)
        # Now we want to translate these indices to next to the corner so that
        # we can use them as transformations to move a voxel to the edge
        final_shell_indices = []
        for a, x in zip(self.shape, shell_indices):
            new_x = x - a
            abs_new_x = np.abs(new_x)
            new_x_filter = abs_new_x < x
            final_x = np.where(new_x_filter, new_x, x)
            final_shell_indices.append(final_x)

        return np.column_stack(final_shell_indices)

    def get_padded_grid_axes(self, padding: int = 0):
        """
        Gets the the possible indices for each dimension of a padded grid.
        e.g. if the original charge density grid is 20x20x20, and is padded
        with one extra layer on each side, this function will return three
        arrays with integers from 0 to 21.

        Args:
            padding (int):
                The amount the grid has been padded

        Returns:
            three arrays with lengths the same as the grids shape
        """
        grid = self.total
        a = np.linspace(
            0,
            grid.shape[0] + (padding - 1) * 2 + 1,
            grid.shape[0] + padding * 2,
        )
        b = np.linspace(
            0,
            grid.shape[1] + (padding - 1) * 2 + 1,
            grid.shape[1] + padding * 2,
        )
        c = np.linspace(
            0,
            grid.shape[2] + (padding - 1) * 2 + 1,
            grid.shape[2] + padding * 2,
        )
        return a, b, c

    def copy(self):
        """
        Convenience method to get a copy of the grid.

        Returns:
            A copy of the Grid.
        """
        return self.__class__(
            self.structure,
            self.data,
        )

    @classmethod
    def from_file(cls, grid_file: str | Path):
        """Create a grid instance using a CHGCAR or ELFCAR file

        Args:
            grid_file (string):
                The file the instance should be made from. Should be a VASP
                CHGCAR or ELFCAR type file.

        Returns:
            Grid from the specified file.
        """
        logging.info(f"Loading {grid_file} from file")
        # Create string to add structure to.
        poscar, data, _ = cls.parse_file(grid_file)

        return Grid(poscar.structure, data)

    def to_pybader(self):
        """
        Returns a Bader object from pybader.
        """
        atoms = self.structure.cart_coords
        lattice = self.matrix
        density = {"charge": self.total}
        file_info = {"voxel_offset": np.array([0, 0, 0])}
        bader = Bader(density, lattice, atoms, file_info)
        return bader

    def run_pybader(self, threads: int = 1):
        """
        Convenience class for running zero-flux voxel assignment using pybader.
        Returns a pybader Bader class object with the assigned voxels

        Args:
            cores (int):
                The number of threads to use when running the Bader algorithm
        """
        logging.info("Running Bader")
        bader = self.to_pybader()
        bader.load_config("speed")
        bader.threads = threads
        bader.spin_flag = True  # loading speed config resets all config vars
        bader.volumes_init()
        bader.bader_calc()
        bader.bader_to_atom_distance()
        bader.refine_volumes(bader.atoms_volumes)
        bader.threads = 1
        bader.min_surface_distance()
        return bader

    def regrid(
        self,
        desired_resolution: int = 130000,
        new_shape: np.array = None,
    ):
        """
        Returns a new grid resized using Fourier interplation as implemented by
        [PyRho](https://materialsproject.github.io/pyrho/)

        Args:
            desired_resolution (int):
                The desired resolution in voxels/A^3.
            new_shape (ArrayLike):
                The new array shape. Takes precedence over desired_resolution.

        Returns:
            Changes the grid data in place.
        """
        # Get data
        total = self.total
        diff = self.diff

        # Get the lattice unit vectors as a 3x3 array
        lattice_array = self.matrix

        # get the original grid size and lattice volume.
        shape = self.shape
        volume = self.structure.volume

        if new_shape is None:
            # calculate how much the number of voxels along each unit cell must be
            # multiplied to reach the desired resolution.
            scale_factor = ((desired_resolution * volume) / shape.prod()) ** (1 / 3)

            # calculate the new grid shape. round up to the nearest integer for each
            # side
            new_shape = np.around(shape * scale_factor).astype(np.int32)

        # create a pyrho pgrid instance.
        total_pgrid = PGrid(total, lattice_array)

        # regrid to the desired voxel resolution and get the data back out.
        new_total_pgrid = total_pgrid.get_transformed(
            [[1, 0, 0], [0, 1, 0], [0, 0, 1]], new_shape
        )
        # new_data = pgrid.lossy_smooth_compression(new_shape)
        new_total_data = new_total_pgrid.grid_data

        # repeat for diff if exists
        if diff is not None:
            diff_pgrid = PGrid(diff, lattice_array)
            new_diff_pgrid = diff_pgrid.get_transformed(
                [[1, 0, 0], [0, 1, 0], [0, 0, 1]], new_shape
            )
            new_diff_data = new_diff_pgrid.grid_data

        data = {"total": new_total_data, "diff": new_diff_data}

        return Grid(self.structure, data)

    ###########################################################################
    # The following is a series of methods that are useful for converting between
    # voxel coordinates, fractional coordinates, and cartesian coordinates.
    # Voxel coordinates go from 0 to grid_size-1. Fractional coordinates go
    # from 0 to 1. Cartesian coordinates convert to real space based on the
    # crystal lattice.
    ###########################################################################
    def get_voxel_coords_from_index(self, site):
        """
        Takes in a site index and returns the equivalent voxel grid index.

        Args:
            site (int):
                the index of the site to find the grid index for

        Returns:
            A voxel grid index as an array.

        """

        voxel_coords = [a * b for a, b in zip(self.shape, self.frac_coords[site])]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_voxel_coords_from_neigh_CrystalNN(self, neigh):
        """
        Gets the voxel grid index from a neighbor atom object from CrystalNN or
        VoronoiNN

        Args:
            neigh (Neigh):
                a neighbor type object from pymatgen

        Returns:
            A voxel grid index as an array.
        """
        grid_size = self.shape
        frac = neigh["site"].frac_coords
        voxel_coords = [a * b for a, b in zip(grid_size, frac)]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_voxel_coords_from_neigh(self, neigh):
        """
        Gets the voxel grid index from a neighbor atom object from the pymatgen
        structure.get_neighbors class.

        Args:
            neigh (dict):
                a neighbor dictionary from pymatgens structure.get_neighbors
                method.

        Returns:
            A voxel grid index as an array.
        """
        grid_size = self.shape
        frac_coords = neigh.frac_coords
        voxel_coords = [a * b for a, b in zip(grid_size, frac_coords)]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_voxel_coords_from_frac(self, frac_coords: ArrayLike | list):
        """
        Takes in a fractional coordinate and returns the cartesian coordinate.

        Args:
            frac_coords (ArrayLike):
                The fractional position to convert to cartesian coords.

        Returns:
            A voxel grid index as an array.

        """
        grid_size = self.shape
        voxel_coords = [a * b for a, b in zip(grid_size, frac_coords)]
        # voxel positions go from 1 to (grid_size + 0.9999)
        return np.array(voxel_coords)

    def get_frac_coords_from_vox(self, voxel_coords: ArrayLike | list):
        """
        Takes in a voxel grid index and returns the fractional
        coordinates.

        Args:
            voxel_coords (ArrayLike):
                A voxel grid index

        Returns:
            A fractional coordinate as an array
        """
        size = self.shape
        fa, fb, fc = [(a / b) for a, b in zip(voxel_coords, size)]
        frac_coords = np.array([fa, fb, fc])
        return frac_coords

    def get_cart_coords_from_frac(self, frac_coords: ArrayLike | list):
        """
        Takes in fractional coordinates and returns cartesian coordinates

        Args:
            frac_coords (ArrayLike):
                The fractional position to convert to cartesian coords.

        Returns:
            Cartesian coordinates as an array
        """
        fa, fb, fc = frac_coords[0], frac_coords[1], frac_coords[2]
        a, b, c = self.a, self.b, self.c
        x = fa * a[0] + fb * b[0] + fc * c[0]
        y = fa * a[1] + fb * b[1] + fc * c[1]
        z = fa * a[2] + fb * b[2] + fc * c[2]
        cart_coords = np.array([x, y, z])
        return cart_coords

    def get_frac_coords_from_cart(self, cart_coords: ArrayLike | list):
        """
        Takes in a cartesian coordinate and returns the fractional coordinates.

        Args:
            cart_coords (ArrayLike):
                A cartesian coordinate.

        Returns:
            fractional coordinates as an Array
        """
        lattice_matrix = self.matrix
        inverse_matrix = np.linalg.inv(lattice_matrix)
        frac_coords = np.dot(cart_coords, inverse_matrix)

        return frac_coords

    def get_cart_coords_from_vox(self, voxel_coords: ArrayLike | list):
        """
        Takes in a voxel grid index and returns the cartesian coordinates.

        Args:
            voxel_coords (ArrayLike):
                A voxel grid index

        Returns:
            Cartesian coordinates as an array
        """
        frac_coords = self.get_frac_coords_from_vox(voxel_coords)
        cart_coords = self.get_cart_coords_from_frac(frac_coords)
        return cart_coords

    def get_voxel_coords_from_cart(self, cart_coords: ArrayLike | list):
        """
        Takes in a cartesian coordinate and returns the voxel coordinates.

        Args:
            cart_coords (ArrayLike): A cartesian coordinate.

        Returns:
            Voxel coordinates as an Array
        """
        frac_coords = self.get_frac_coords_from_cart(cart_coords)
        voxel_coords = self.get_voxel_coords_from_frac(frac_coords)
        return voxel_coords

    def get_cart_coords_from_frac_full_array(self, frac_coords: ArrayLike):
        """
        Takes in a 2D array of shape (N,3) representing fractional coordinates
        at N points and calculates the equivalent cartesian coordinates.

        Args:
            frac_coords (ArrayLike):
                An (N,3) shaped array of fractional coordinates

        Returns:
            An (N,3) shaped array of cartesian coordinates
        """
        x, y, z = self.matrix.T
        cart_x = np.dot(frac_coords, x)
        cart_y = np.dot(frac_coords, y)
        cart_z = np.dot(frac_coords, z)
        cart_coords = np.column_stack([cart_x, cart_y, cart_z])
        return cart_coords

    def get_frac_coords_from_vox_full_array(self, vox_coords: ArrayLike):
        """
        Takes in a 2D array of shape (N,3) representing voxel coordinates
        at N points and calculates the equivalent fractional coordinates.

        Args:
            vox_coords (ArrayLike):
                An (N,3) shaped array of voxel coordinates

        Returns:
            An (N,3) shaped array of fractional coordinates
        """
        x, y, z = self.shape
        frac_x = vox_coords[:, 0] / x
        frac_y = vox_coords[:, 1] / y
        frac_z = vox_coords[:, 2] / z
        frac_coords = np.column_stack([frac_x, frac_y, frac_z])
        return frac_coords

    def get_vox_coords_from_frac_full_array(self, frac_coords: ArrayLike):
        """
        Takes in a 2D array of shape (N,3) representing fractional coordinates
        at N points and calculates the equivalent voxel coordinates.

        Args:
            frac_coords (ArrayLike):
                An (N,3) shaped array of fractional coordinates

        Returns:
            An (N,3) shaped array of voxel coordinates
        """
        x, y, z = self.shape
        vox_x = frac_coords[:, 0] * x
        vox_y = frac_coords[:, 1] * y
        vox_z = frac_coords[:, 2] * z
        vox_coords = np.column_stack([vox_x, vox_y, vox_z])
        return vox_coords

    def get_cart_coords_from_vox_full_array(self, vox_coords: ArrayLike):
        """
        Takes in a 2D array of shape (N,3) representing voxel coordinates
        at N points and calculates the equivalent cartesian coordinates.

        Args:
            frac_coords (ArrayLike):
                An (N,3) shaped array of voxel coordinates

        Returns:
            An (N,3) shaped array of cartesian coordinates
        """
        frac_coords = self.get_frac_coords_from_vox_full_array(vox_coords)
        return self.get_cart_coords_from_frac_full_array(frac_coords)

    def _plot_points(self, points, ax, fig, color, size: int = 20):
        """
        Plots points of form [x,y,z] using matplotlib

        Args:
            points (list): A list of points to plot
            fig: A matplotlib.pyplot.figure() instance
            ax: A matplotlib Subplot instance
            color (str): The color to plot the points
            size (int): The pt size to plot
        """
        x = []
        y = []
        z = []
        for point in points:
            x.append(point[0])
            y.append(point[1])
            z.append(point[2])
        ax.scatter(x, y, z, c=color, s=size)

    def _plot_unit_cell(self, ax, fig):
        """
        Plots the unit cell of a structure using matplotlib

        Args:
            fig: A matplotlib.pyplot.figure() instance
            ax: A matplotlib Subplot instance
        """
        if ax is None or fig is None:
            fig = plt.figure()
            ax = fig.add_subplot(projection="3d")

        # Get the points at the lattice vertices to plot and plot them
        a = self.a
        b = self.b
        c = self.c
        points = [np.array([0, 0, 0]), a, b, c, a + b, a + c, b + c, a + b + c]
        self._plot_points(points, ax, fig, "purple")

        # get the structure to plot.
        structure = self.structure
        species = structure.symbol_set

        # get a color map to distinguish between sites
        color_map = matplotlib.colormaps.get_cmap("tab10")
        # Go through each atom type and plot all instances with the same color
        for i, specie in enumerate(species):
            color = color_map(i)
            sites_indices = structure.indices_from_symbol(specie)
            for site in sites_indices:
                coords = structure[site].coords
                self._plot_points([coords], ax, fig, color, size=40)



# -*- coding: utf-8 -*-

import csv
import logging
import math
import warnings
from functools import cached_property
from pathlib import Path

import numpy as np
import psutil
from pymatgen.analysis.local_env import CrystalNN
from pymatgen.io.vasp import Potcar
from scipy.ndimage import label
from tqdm import tqdm

from simmate.apps.badelf.core.electride_finder import ElectrideFinder
from simmate.apps.badelf.core.grid import Grid
from simmate.apps.badelf.core.partitioning import PartitioningToolkit
from simmate.apps.badelf.core.voxel_assignment import VoxelAssignmentToolkit


class BadElfToolkit:
    """
    A set of tools for performing BadELF, VoronELF, or Zero-Flux analysis on
    outputs from a VASP calculation.

    Args:
        partitioning_grid (Grid):
            A badelf app Grid like object used for partitioning the unit cell
            volume. Usually contains ELF.
        charge_grid (Grid):
            A badelf app Grid like object used for summing charge. Usually
            contains charge density.
        directory (Path):
            The Path to perform the analysis in.
        threads (int):
            The number of threads to use for voxel assignment.
            Defaults to 0.9*the total number of threads available.
        algorithm (str):
            The algorithm to use for partitioning. Options are "badelf", "voronelf",
            or "zero-flux".
        find_electrides (bool):
            Whether or not to search for electride sites. Usually set to true.
    """

    check_for_covalency = True
    electride_finder_cutoff = 0.5

    def __init__(
        self,
        partitioning_grid: Grid,
        charge_grid: Grid,
        directory: Path,
        threads: int = None,
        algorithm: str = "badelf",
        find_electrides: bool = True,
    ):
        if partitioning_grid.structure != charge_grid.structure:
            raise ValueError("Grid structures must be the same.")
        if threads is None:
            self.threads = math.floor(psutil.Process().num_threads() * 0.9 / 2)
        else:
            self.threads = threads

        # Check if POTCAR exists in path. If not, throw warning
        if not (directory / "POTCAR").exists():
            raise Exception(
                """
                No POTCAR file found in the requested directory.
                """
            )
        if algorithm not in ["badelf", "voronelf", "zero-flux"]:
            raise ValueError(
                """The algorithm setting you chose does not exist. Please select
                  either 'badelf', 'voronelf', or 'zero-flux'.
                  """
            )

        self.partitioning_grid = partitioning_grid
        self.charge_grid = charge_grid
        self.directory = directory
        self.algorithm = algorithm
        self.find_electrides = find_electrides

    @cached_property
    def structure(self):
        structure = self.partitioning_grid.structure.copy()
        structure.remove_species(["He"])
        return structure

    @cached_property
    def electride_structure(self):
        """
        Searches the partitioning grid for potential electride sites and returns
        a structure with the found sites.

        Returns:
            A Structure object with electride sites as "He" atoms.
        """

        if self.find_electrides:
            electride_structure = ElectrideFinder(
                self.partitioning_grid.copy(),
                self.directory,
            ).get_electride_structure(
                electride_finder_cutoff=self.electride_finder_cutoff,
                check_for_covalency=self.check_for_covalency,
                threads=self.threads,
            )
        else:
            electride_structure = self.structure

        return electride_structure

    @cached_property
    def electride_indices(self):
        """
        The indices of the structure that are electride sites.
        """
        return np.array(list(self.electride_structure.indices_from_symbol("He")))

    @cached_property
    def coord_envs(self):
        """
        The coordination environment around each electride.
        """
        return self._get_coord_envs()

    def _get_coord_envs(self):
        """
        Gets the coordination environment for electrides in the system.

        Returns:
            A list of coordination environments in order of the electrides in
            the system.
        """
        # create a CrystalNN loop
        cnn = CrystalNN()
        coord_envs = []
        # For each electride in the structure, we make a new temporary structure
        # without electride sites. We then add one electride site at a time and
        # check its environment.
        for i, site in enumerate(self.electride_structure):
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", category=UserWarning, module="pymatgen"
                )
                coord_envs.append(cnn.get_cn(structure=self.electride_structure, n=i))
        return coord_envs

    @cached_property
    def partitioning(self):
        """
        The partitioning planes for the structure as a dictionary of dataframes.
        None if the zero-flux method is selected
        """
        return self._get_partitioning()

    def _get_partitioning(self):
        """
        Gets the partitioning used in the badelf and voronelf algorithms.

        Returns:
            Dictionary relating sites to the planes surrounding them. None
            if the zero-flux algorithm is selected.
        """
        # Get the partitioning grid
        partitioning_grid = self.partitioning_grid.copy()
        # If the algorithm is badelf, we don't want to partition with the structure
        # containing electrides. We remove any electrides in case the provided
        # structure already had them.
        if self.algorithm == "badelf":
            # remove electrides from grid structure and get
            partitioning_grid.structure.remove_species("He")
            partitioning = PartitioningToolkit(
                partitioning_grid, self.pybader
            ).get_partitioning()
            return partitioning
        elif self.algorithm == "voronelf":
            # Use the structure with electrides as the partitioning structure.
            # This will not be anything different from the base structure if there
            # are no electride sites.
            partitioning_grid.structure = self.electride_structure.copy()
            partitioning = PartitioningToolkit(
                partitioning_grid, self.pybader
            ).get_partitioning()
            return partitioning
        elif self.algorithm == "zero-flux":
            print(
                """
                There is no partitioning property for the zero-flux algorithm as
                the partitioning is handled by the [Henkelman Bader code](https://theory.cm.utexas.edu/henkelman/code/bader/)
                """
            )
            return None

    @property
    def single_site_voxel_assignments(self):
        """
        A 1D array with site assignments for voxels not exactly on a partitioning
        plane
        """
        return self.voxel_assignments[0]

    @property
    def multi_site_voxel_assignments(self):
        """
        An (N,M) shaped array with indices i,j where i is the voxel index
        (these are sub indices, full indices are stored in multi_site_voxel_indices)
        and j is the site. A 1 indicates that this voxel is partially shared by
        this site
        """
        return self.voxel_assignments[1]

    @property
    def multi_site_voxel_indices(self):
        """
        The corresponding voxel indices for the multi_site_voxel_assignments
        array.
        """
        return np.where(self.single_site_voxel_assignments == 0)[0]

    @cached_property
    def voxel_assignments_array(self):
        """
        A 3D array with the same shape as the charge grid indicating where
        voxels are assigned
        """
        # Get multi site indices
        split_voxel_indices = self.multi_site_voxel_indices
        # create a list to store the randomly assigned voxels
        random_voxel_assignments = []
        # loop through the voxels belonging to multiple sites to pick one random
        # site to assign to.
        for split_voxel in self.multi_site_voxel_assignments:
            # round the assignments
            split_voxel = np.round(split_voxel, 2)
            if len(np.unique(split_voxel)) == 2:
                # There are several sites that share an equal part of this voxel.
                # We want to randomly assign to one of these
                # Get which sites this voxel is split by
                possible_sites = np.where(split_voxel != 0)[0]
                # Pick one randomly
                # site_choice = np.random.choice(possible_sites)
                # !!! For the purposes of electride dimensionality it is better to
                # assign shared voxels to electrides which always have higher index
                # values.
                # !!! With new assignment method it may be better to assign to
                # highest frac or random
                site_choice = possible_sites.max()
                # append it to our list
                random_voxel_assignments.append(site_choice)
            else:
                # There are multiple ratios. We want to assign to the site with
                # the most
                site_choice = np.where(split_voxel == np.max(split_voxel))[0][0]
                random_voxel_assignments.append(site_choice)
        # Get the single site assignments and subtract one to get to sites
        # beginning at 0
        all_site_assignments = self.single_site_voxel_assignments.copy() - 1
        # Assign our randomly generated sites then return the array as a 3D grid
        all_site_assignments[split_voxel_indices] = np.array(random_voxel_assignments)
        return all_site_assignments.reshape(self.charge_grid.shape)

    @cached_property
    def pybader(self):
        """
        A pybader Bader object run on the PartitioningGrid
        """
        bader_grid = self.partitioning_grid.copy()
        bader_grid.structure = self.electride_structure
        bader = bader_grid.run_pybader(threads=self.threads)
        return bader

    @cached_property
    def zero_flux_voxel_assignments(self):
        """
        An array of the same size as the partitioning grid representing the site
        assignments from the zero-flux partitioning method. This uses the
        [pybader code](https://github.com/adam-kerrigan/pybader)
        """
        return self.pybader.atoms_volumes.copy()

    @cached_property
    def voxel_assignments(self):
        """
        Two arrays representing voxels assigned to only one site and voxels
        assigned to multiple sites
        """
        return self._get_voxel_assignments()

    def _get_voxel_assignments(self):
        """
        Gets a dataframe of voxel assignments. The dataframe has columns
        [x, y, z, charge, sites]
        """
        logging.info("Beginning voxel assignment (this can take a while)")
        algorithm = self.algorithm
        # Get the zero-flux voxel assignments
        all_voxel_site_assignments = self.zero_flux_voxel_assignments.ravel()
        # shift voxel assignments to convention where 0 is unassigned
        all_voxel_site_assignments += 1
        if algorithm == "zero-flux":
            single_site_voxel_assignments = all_voxel_site_assignments
            multi_site_voxel_assignments = np.array([])

        # Get the objects that we'll need to assign voxels.
        elif algorithm in ["badelf", "voronelf"]:
            charge_grid = self.charge_grid
            partitioning = self.partitioning
            voxel_assignment_tools = VoxelAssignmentToolkit(
                charge_grid=charge_grid,
                electride_structure=self.electride_structure,
                partitioning=partitioning,
                algorithm=self.algorithm,
                directory=self.directory,
            )
            if algorithm == "badelf":
                # Get the voxel assignments for each electride site
                all_voxel_site_assignments = np.where(
                    np.isin(all_voxel_site_assignments, self.electride_indices + 1),
                    all_voxel_site_assignments,
                    0,
                )
            else:
                # get an initial array of no site assignments
                all_voxel_site_assignments = np.zeros(charge_grid.voxel_num)
            # get assignments for voxels belonging to single sites
            single_site_voxel_assignments = (
                voxel_assignment_tools.get_single_site_voxel_assignments(
                    all_voxel_site_assignments
                )
            )
            # get assignments for voxels split by two or more sites
            multi_site_voxel_assignments = (
                voxel_assignment_tools.get_multi_site_voxel_assignments(
                    single_site_voxel_assignments.copy()
                )
            )

        logging.info("Finished voxel assignment")
        return (
            single_site_voxel_assignments,
            multi_site_voxel_assignments,
        )

    @staticmethod
    def get_ELF_dimensionality(grid: Grid, cutoff: float):
        """
        This algorithm works by checking if the voxels with values above the cutoff
        are connected to the equivalent voxel in the unit cell one transformation
        over. This is done primarily using scipy.ndimage.label which determines
        which voxels are connected. To do this rigorously, the unit cell is repeated
        to make a (2,2,2) super cell and the connections are checked going from
        the original unit cell to the unit cells connected at the faces, edges,
        and corners. If a connection in that direction is found, the total number
        of connections increases. Dimensionalities of 0,1,2, and 3 are represented
        by 0,1,4,and 13 connections respectively.

        Args:
            cutoff (float):
                The minimum elf value to consider as a connection.
            grid (Grid):
                The ELF Grid object with only values associated with electrides.
        """
        data = grid.total
        voxel_indices = np.indices(grid.shape).reshape(3, -1).T
        # Remove data below our cutoff
        thresholded_data = np.where(data <= cutoff, 0, 1)
        raveled_data = thresholded_data.ravel()
        # Label connected components in the eroded data. We will check each distinct
        # body to see if it connects in each direction
        structure = np.ones([3, 3, 3])
        labels, num_features = label(thresholded_data, structure=structure)
        if num_features == 0:
            return 0
        # We need to get one voxel in each of the features that we can transpose
        # later on
        feature_indices = []
        for i, site in enumerate(grid.structure):
            if site.species_string == "He":
                frac_coords = site.frac_coords
                voxel_coords = grid.get_voxel_coords_from_frac(frac_coords).astype(int)
                # Make sure that the label is not 0
                site_label = labels[voxel_coords[0], voxel_coords[1], voxel_coords[2]]
                if site_label != 0:
                    feature_indices.append(voxel_coords)
        if len(feature_indices) == 0:
            return 0

        # We are going to need to translate the above voxels and the entire unit
        # cell so we create a list of desired transformations
        transformations = [
            [0, 0, 0],  # -
            [1, 0, 0],  # x
            [0, 1, 0],  # y
            [0, 0, 1],  # z
            [1, 1, 0],  # xy
            [1, 0, 1],  # xz
            [0, 1, 1],  # yz
            [1, 1, 1],  # xyz
        ]
        transformations = np.array(transformations)
        transformations = grid.get_voxel_coords_from_frac(transformations.T).T
        # create a new numpy array representing the data tiled once in each direction
        # to get a corner
        supercell_data = np.zeros(np.array(thresholded_data.shape) * 2)
        for transformation in transformations:
            transformed_indices = (voxel_indices + transformation).astype(int)
            x = transformed_indices[:, 0]
            y = transformed_indices[:, 1]
            z = transformed_indices[:, 2]
            supercell_data[x, y, z] = raveled_data
        # convert data into labeled features
        supercell_data, _ = label(supercell_data, structure)
        # The unit cell can be connected to neighboring unit cells in 26 directions.
        # however, we only need to consider half of these as the others are symmetrical.
        connections = [
            # surfaces (3)
            [0, 1],  # x
            [0, 2],  # y
            [0, 3],  # z
            # edges (6)
            [0, 4],  # xy
            [0, 5],  # xz
            [0, 6],  # yz
            [3, 1],  # x-z
            [3, 2],  # y-z
            [1, 2],  # -xy
            # corners (4)
            [0, 7],  # x,y,z
            [1, 6],  # -x,y,z
            [2, 5],  # x,-y,z
            [3, 4],  # x,y,-z
        ]
        # Using these connections we can determine the dimensionality of the system.
        # 1 connection is 1D, 2-4 connections is 2D and 5-13 connections is 3D.
        # !!! These may need to be updated if I'm wrong. The idea comes from
        # the fact that the connections should be 1, 4, and 13, but sometimes
        # voxelation issues result in a connection not working in one direction
        # while it would in the reverse direction (which isn't possible with
        # true symmetry). The range accounts for this possibility. The problem
        # might be if its possible to have for example a 2D connecting structure
        # with 5 connections. However, I'm pretty sure that immediately causes
        # an increase to 3D dimensionality.
        # First we create a list to store potential dimensionalites based off of
        # each feature. We will take the highest dimensionality.
        dimensionalities = []
        for coord in feature_indices:
            # create count for the number of connections
            connections_num = 0
            for connection in connections:
                # Get voxel coordinates at the first transformation
                x, y, z = (coord + transformations[connection[0]]).astype(int)
                # get voxel coordinates at the second transformation
                x1, y1, z1 = (coord + transformations[connection[1]]).astype(int)
                # get the feature label at each voxel
                label1 = supercell_data[x, y, z]
                label2 = supercell_data[x1, y1, z1]
                # If the labels are the same, the unit cell is connected in this
                # direction
                if label1 == label2:
                    connections_num += 1
            if connections_num == 0:
                dimensionalities.append(0)
            elif connections_num == 1:
                dimensionalities.append(1)
            elif 1 < connections_num <= 4:
                dimensionalities.append(2)
            elif 5 < connections_num <= 13:
                dimensionalities.append(3)

        return max(dimensionalities)

    def get_electride_dimensionality(self):
        """
        Finds the dimensionality (e.g. 0D, 1D, etc.) of an electride by labeling
        features in the ELF at various cutoffs and determining if they are
        connected to nearby unit cells.

        Args:
            grid (Grid):
                The ELF Grid object with only values associated with electrides.
        """
        electride_indices = self.electride_indices
        # If we have no electrides theres no reason to continue so we stop here
        if len(electride_indices) == 0:
            return None, None

        ###############################################################################
        # This section preps an ELF grid that only contains values from the electride
        # sites and is zero everywhere else.
        ###############################################################################

        # read in ELF data and regrid so that it is the same size as the
        # charge grid
        elf_grid = self.partitioning_grid.copy()
        # elf_grid = elf_grid.regrid(desired_resolution=self.charge_grid.voxel_resolution)
        voxel_assignment_array = self.voxel_assignments_array
        # Get array where values are ELF values when voxels belong to electrides
        # and are 0 otherwise
        elf_array = np.where(
            np.isin(voxel_assignment_array, electride_indices), elf_grid.total, 0
        )
        elf_grid.total = elf_array
        elf_grid.structure = self.electride_structure

        #######################################################################
        # This section scans across different cutoffs to determine what dimensionalities
        # exist in the electride ELF
        #######################################################################
        logging.info("Finding dimensionality cutoffs")
        logging.info("Calculating dimensionality at 0 ELF")
        highest_dimension = self.get_ELF_dimensionality(elf_grid, 0)
        dimensions = [i for i in range(0, highest_dimension)]
        dimensions.reverse()
        # Create lists for the refined dimensions
        final_dimensions = [highest_dimension]
        final_connections = [0]
        amounts_to_change = []
        # We refine by guessing the cutoff is 0.5 then increasing or decreasing by
        # 0.25, then 0.125 etc. down to 0.000015259.
        for i in range(1, 16):
            amounts_to_change.append(1 / (2 ** (i + 1)))
        for dimension in dimensions:
            guess = 0.5
            # assume this dimension is not found
            found_dimension = False
            logging.info(f"Refining cutoff for dimension {dimension}")
            for i in tqdm(amounts_to_change, total=len(amounts_to_change)):
                # check what our current dimension is. If we are at a higher dimension
                # we need to raise the cutoff. If we are at a lower dimension or at
                # the dimension we need to lower it
                current_dimension = self.get_ELF_dimensionality(elf_grid, guess)
                if current_dimension > dimension:
                    guess += i
                elif current_dimension < dimension:
                    guess -= i
                elif current_dimension == dimension:
                    # We have found the dimension so we add it to our lists.
                    guess -= i
                    found_dimension = True
            if found_dimension:
                final_connections.append(round(guess, 4))
                final_dimensions.append(dimension)
        return final_dimensions, final_connections

    @cached_property
    def results(self):
        """
        A summary of the results from a BadELF run.
        """
        return self._get_results()

    def _get_results(self):
        """
        Gets the results for a BadELF run and prints them to a .csv file.
        """
        algorithm = self.algorithm
        directory = self.directory
        electride_num = len(self.electride_indices)
        electride_structure = self.electride_structure
        electride_indices = self.electride_indices
        a, b, c = self.charge_grid.shape
        elements = []
        for site in electride_structure:
            if site.species_string == "He":
                elements.append("e")
            else:
                elements.append(site.species_string)

        # get the voxel assignments. Note that the convention here is to
        # have indices starting at 1 rather than 0
        (
            single_site_assignments,
            multi_site_assignments,
        ) = self.voxel_assignments
        voxel_volume = self.charge_grid.voxel_volume
        charge_array = self.charge_grid.total.ravel()
        logging.info("Calculating additional useful information")
        # Create dictionaries to store the important results
        min_dists = {}
        charges = {}
        atomic_volumes = {}
        # Get min dists
        for site in range(len(electride_structure)):
            charges[site] = 0
            atomic_volumes[site] = 0
        # Get the minimum distances from each atom the the partitioning
        # surface. If the algorithm is "badelf" we need to acquire the
        # radii for the electrides
        if algorithm == "badelf" and len(electride_indices > 0):
            bader = self.pybader
            distances = bader.atoms_surface_distance
            electride_min_dists = distances[self.electride_indices]

        if algorithm == "zero-flux":
            bader = self.pybader
            distances = bader.atoms_surface_distance
            for i, distance in enumerate(distances):
                min_dists[i] = distance
        else:
            for site in range(len(electride_structure)):
                # fill min_dist dictionary using the smallest partitioning radius
                if site in electride_indices and algorithm == "badelf":
                    # Get dist from henkelman algorithm results
                    min_dists[site] = electride_min_dists[site - len(self.structure)]
                else:
                    # Get dists from partitioning
                    radii = []
                    partitioning = self.partitioning
                    for neighbor_index, row in partitioning[site].iterrows():
                        radii.append(row["radius"])
                    min_radii = min(radii)
                    min_dists[site] = min_radii

        # Get the charge and atomic volume of each site for sites with
        # one assignment
        for site in range(len(electride_structure)):
            site1 = site + 1
            voxel_indices = np.where(single_site_assignments == site1)[0]
            site_charge = charge_array[voxel_indices]
            charges[site] += np.sum(site_charge)
            atomic_volumes[site] += len(voxel_indices) * voxel_volume

        # Get the charge and atomic volume of each voxel with multiple site
        # assignments. These are stored as a (N,M) shaped array where N
        # is the number of split voxels and M is the number of atoms.
        if len(self.multi_site_voxel_assignments) > 0:
            split_voxel_indices = self.multi_site_voxel_indices
            split_voxel_charge = charge_array[split_voxel_indices]
            # get how many sites each voxel is split by
            # split_voxel_ratio = 1 / np.sum(multi_site_assignments, axis=1)
            for site_index, assignment_array in enumerate(multi_site_assignments.T):
                indices_where_assigned = np.where(assignment_array > 0)[0]
                site_charge = split_voxel_charge[indices_where_assigned]
                site_charge = site_charge * assignment_array[indices_where_assigned]
                charges[site_index] += np.sum(site_charge)
                atomic_volumes[site] += (
                    np.sum(assignment_array[indices_where_assigned]) * voxel_volume
                )

        # Convert charges from VASP standard
        for site, charge in charges.items():
            charges[site] = round((charge / (a * b * c)), 4)
        for site, volume in atomic_volumes.items():
            atomic_volumes[site] = round(volume, 4)

        # Get the number of electrons assigned by badelf.
        nelectrons = round(sum(charges.values()), 4)

        # Get
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            potcars = Potcar.from_file(directory / "POTCAR")
        nelectron_data = {}
        # the result is a list because there can be multiple element potcars
        # in the file (e.g. for NaCl, POTCAR = POTCAR_Na + POTCAR_Cl)
        for potcar in potcars:
            nelectron_data.update({potcar.element: potcar.nelectrons})

        # create lists to store the element list, oxidation states, charges,
        # minimum distances, and atomic volumes
        oxi_state_data = []
        charges_list = []
        min_dists_list = []
        atomic_volumes_list = []
        # iterate over the charge results and add the results to each list
        for site_index, site_charge in charges.items():
            # get structure site
            site = electride_structure[site_index]
            # get element name
            element_str = site.specie.name
            # Change electride dummy atom name to e
            if element_str == "He":
                element_str = "e"
            # calculate oxidation state and add it to the oxidation state list
            if element_str == "e":
                oxi_state = -site_charge
            else:
                oxi_state = round((nelectron_data[element_str] - site_charge), 4)
            oxi_state_data.append(oxi_state)
            # add the corresponding charge, distance, and atomic volume to the
            # respective lits
            charges_list.append(site_charge)
            min_dists_list.append(round(min_dists[site_index], 4))
            atomic_volumes_list.append(atomic_volumes[site_index])

        # Calculate the "vacuum charge" or the charge not associated with any atom.
        # Idealy this should be 0.
        total_electrons = 0
        for site in self.structure:
            site_valence_electrons = nelectron_data[site.species_string]
            total_electrons += site_valence_electrons
        vacuum_charge = round((total_electrons - nelectrons), 4)

        # Calculate the "vacuum volume" or the volume not associated with any atom.
        # Idealy this should be 0
        vacuum_volume = round((self.structure.volume - sum(atomic_volumes.values())), 4)

        # Save everything in a results dictionary
        results = {
            "oxidation_states": oxi_state_data,
            "charges": charges_list,
            "min_dists": min_dists_list,
            "atomic_volumes": atomic_volumes_list,
            "vacuum_charge": vacuum_charge,
            "vacuum_volume": vacuum_volume,
            "nelectrons": nelectrons,
        }
        # calculate the number of electride electrons per formula unit
        if len(self.electride_indices) > 0:
            electrides_per_unit = sum(
                [results["charges"][i] for i in self.electride_indices]
            )
            (
                _,
                formula_reduction_factor,
            ) = self.structure.composition.get_reduced_composition_and_factor()
            electrides_per_reduced_unit = electrides_per_unit / formula_reduction_factor
            results["electrides_per_formula"] = electrides_per_unit
            results["electrides_per_reduced_formula"] = electrides_per_reduced_unit
        # set the results that are not algorithm dependent
        results["nelectrides"] = electride_num
        results["algorithm"] = algorithm
        results["element_list"] = elements
        results["coord_envs"] = self.coord_envs
        electride_dim, dim_cutoffs = self.get_electride_dimensionality()
        results["electride_dim"] = electride_dim
        results["dim_cutoffs"] = dim_cutoffs
        # Fill out columns unrelated to badelf alg
        structure = self.structure
        results["structure"] = structure

        return results

    def write_results_csv(self):
        directory = self.directory
        results = self.results
        with open(directory / "badelf_summary.csv", "w") as csv_file:
            writer = csv.writer(csv_file)
            for key, value in results.items():
                writer.writerow([key, value])

    @classmethod
    def from_files(
        cls,
        directory: Path = Path("."),
        partitioning_file: str = "ELFCAR",
        charge_file: str = "CHGCAR",
        algorithm: str = "badelf",
        find_electrides: bool = True,
    ):
        """
        Creates a BadElfToolkit instance from the requested partitioning file
        and charge file.

        Args:
            directory (Path):
                The path to the directory that the badelf analysis
                will be located in.
            partitioning_file (str):
                The filename of the file to use for partitioning. Must be a VASP
                CHGCAR or ELFCAR type file.
            charge_file (str):
                The filename of the file containing the charge information. Must
                be a VASP CHGCAR or ELFCAR type file.
            algorithm (str):
                The algorithm to use. Options are "badelf", "voronelf", or "zero-flux"
            find_electrides (bool):
                Whether or not to search for electrides in the system

        Returns:
            A BadElfToolkit instance.
        """

        partitioning_grid = Grid.from_file(directory / partitioning_file)
        charge_grid = Grid.from_file(directory / charge_file)
        return BadElfToolkit(
            partitioning_grid=partitioning_grid,
            charge_grid=charge_grid,
            directory=directory,
            algorithm=algorithm,
            find_electrides=find_electrides,
        )

    def write_species_file(self, file_type: str = "ELFCAR", species: str = "He"):
        """
        Writes an ELFCAR or CHGCAR for a given species. Writes to the default
        directory provided to the BadelfToolkit class.

        Args:
            file_type (str):
                The type of file that you want, either ELFCAR or CHGCAR
            species (str):
                The species to write data for.

        Returns:
            None
        """
        if self.algorithm == "zero-flux":
            warnings.warn(
                """
                There is currently no method to write files for 
                assignments found using the zero-flux method. This method is
                a wrap around for the Henkelman group's bader software. We
                suggest using their software for more involved zero-flux
                workflows.
                """
            )
            return
        # Get directory
        directory = self.directory
        # Get voxel assignments and data
        voxel_assignment_array = self.voxel_assignments_array
        if file_type == "ELFCAR":
            grid = self.partitioning_grid.copy()
            # grid = grid.regrid(desired_resolution=self.charge_grid.voxel_resolution)
        elif file_type == "CHGCAR":
            grid = self.charge_grid.copy()
        else:
            raise ValueError(
                """
                Invalid file_type. Options are "ELFCAR" or "CHGCAR".
                """
            )
        grid.structure = self.electride_structure
        indices = self.electride_structure.indices_from_symbol(species)
        # Get array where values are ELF values when voxels belong to electrides
        # and are 0 otherwise
        array = np.where(np.isin(voxel_assignment_array, indices), grid.total, 0)
        grid.total = array
        if grid.diff is not None:
            diff_array = np.where(
                np.isin(voxel_assignment_array, indices), grid.diff, 0
            )
            grid.diff = diff_array

        if species == "He":
            species = "e"
        if file_type == "ELFCAR":
            grid.write_file(directory / f"ELFCAR_{species}")
        elif file_type == "CHGCAR":
            grid.write_file(directory / f"CHGCAR_{species}")

    def write_atom_file(
        self,
        atom_index: int,
        file_type: str = "ELFCAR",
    ):
        """
        Writes an ELFCAR or CHGCAR for a given atom. Writes to the default
        directory provided to the BadelfToolkit class.

        Args:
            file_type (str):
                The type of file that you want, either ELFCAR or CHGCAR
            atom_index (str):
                The index of the atom to write data for.

        Returns:
            None
        """
        if self.algorithm == "zero-flux":
            warnings.warn(
                """There is currently no method to write files for 
                   assignments found using the zero-flux method. This method is
                   a wrap around for the Henkelman group's bader software. We
                   suggest using their software for more involved zero-flux
                   workflows.
                """
            )
            return
        # Get directory
        directory = self.directory
        # Get voxel assignments and data
        voxel_assignment_array = self.voxel_assignments_array
        if file_type == "ELFCAR":
            grid = self.partitioning_grid.copy()
            # grid.regrid(desired_resolution=self.charge_grid.voxel_resolution)
        elif file_type == "CHGCAR":
            grid = self.charge_grid.copy()
        else:
            raise ValueError(
                """
                Invalid file_type. Options are "ELFCAR" or "CHGCAR".
                """
            )
        grid.structure = self.electride_structure
        # Get array where values are ELF values when voxels belong to electrides
        # and are 0 otherwise
        array = np.where(np.isin(voxel_assignment_array, atom_index), grid.total, 0)
        grid.total = array
        if grid.diff is not None:
            diff_array = np.where(
                np.isin(voxel_assignment_array, atom_index), grid.diff, 0
            )
            grid.diff = diff_array

        if file_type == "ELFCAR":
            grid.write_file(directory / f"ELFCAR_{atom_index}")
        elif file_type == "CHGCAR":
            grid.write_file(directory / f"CHGCAR_{atom_index}")

    def plot_partitioning(self):
        """
        Plots the partitioning surface around each atom.
        """
        partitioning, _ = self.partitioning
        grid = self.partitioning_grid.copy()
        if self.algorithm == "badelf":
            grid.structure = self.structure
            PartitioningToolkit(grid, self.pybader).plot_partitioning_results(
                partitioning
            )
        elif self.algorithm == "voronelf":
            grid.structure = self.electride_structure
            PartitioningToolkit(grid, self.pybader).plot_partitioning_results(
                partitioning
            )
        else:
            warnings.warn(
                """
                Plotting of zero-flux partitioning surfaces is not currently
                supported.
                """
            )



# -*- coding: utf-8 -*-

import itertools
import logging
import math
from functools import cached_property
from pathlib import Path

import dask.array as da
import numpy as np
import psutil
from numpy.typing import ArrayLike
from tqdm import tqdm

from simmate.apps.badelf.core.grid import Grid
from simmate.toolkit import Structure


class VoxelAssignmentToolkit:
    """
    A set of tools for assigning charge to atoms in a unit cell.

    Args:
        charge_grid (Grid):
            A BadELF app Grid type object usually with CHGCAR type data.
        partitioning_grid (Grid):
            A BadELF app Grid type object usually with ELFCAR type data.
        algorithm (str):
            The algorithm to use for partitioning. Defaults to BadELF
        partitioning (dict):
            A partitioning dictionary generated from the BadELF app
            PartitioningToolkit. Will be generated from the grid if None.
        electride_structure (Structure):
            The structure with electride sites. Will be generated if not given.

    """

    def __init__(
        self,
        charge_grid: Grid,
        electride_structure: Structure,
        algorithm: str,
        partitioning: dict,
        directory: Path,
    ):
        self.charge_grid = charge_grid.copy()
        self.algorithm = algorithm
        # partitioning will contain electride sites for voronelf
        self.partitioning = partitioning
        self.electride_structure = electride_structure

    @property
    def unit_cell_permutations_vox(self):
        """
        The permutations required to transform a unit cell to each of its neighbors.
        Uses voxel coordinates.
        """
        return self.charge_grid.permutations

    @property
    def unit_cell_permutations_frac(self):
        """
        The permutations required to transform a unit cell to each of its neighbors.
        Uses fractional coordinates.
        """
        unit_cell_permutations_frac = [
            (t, u, v)
            for t, u, v in itertools.product([-1, 0, 1], [-1, 0, 1], [-1, 0, 1])
        ]
        # move 0,0,0 transformation to front as it is the most likely to be important
        unit_cell_permutations_frac.insert(0, unit_cell_permutations_frac.pop(13))
        return unit_cell_permutations_frac

    @property
    def unit_cell_permutations_cart(self):
        """
        The permutations required to transform a unit cell to each of its neighbors.
        Uses cartesian coordinates.
        """
        grid = self.charge_grid
        return grid.get_cart_coords_from_frac_full_array(
            self.unit_cell_permutations_frac
        )

    @property
    def vertices_transforms_frac(self):
        """
        The transformations required to transform the center of a voxel to its
        corners. Uses fractional coordinates.
        """
        a, b, c = self.charge_grid.grid_shape
        a1, b1, c1 = 1 / (2 * a), 1 / (2 * b), 1 / (2 * c)
        x, y, z = np.meshgrid([-a1, a1], [-b1, b1], [-c1, c1])
        vertices_transforms_frac = np.column_stack((x.ravel(), y.ravel(), z.ravel()))
        return vertices_transforms_frac

    @property
    def vertices_transforms_cart(self):
        """
        The transformations required to transform the center of a voxel to its
        corners. Uses cartesian coordinates.
        """
        return self.charge_grid.get_cart_coords_from_frac_full_array(
            self.vertices_transforms_frac
        )

    @cached_property
    def all_voxel_frac_coords(self):
        """
        The fractional coordinates for all of the voxels in the charge grid
        """
        return self.charge_grid.all_voxel_frac_coords

    @cached_property
    def all_partitioning_plane_points_and_vectors(self):
        """
        The points and vectors for all of the partitioning planes stored as
        two sets of N,3 shaped arrays.
        """
        partitioning = self.partitioning
        plane_points = []
        plane_vectors = []
        for atom_index, partitioning_df in partitioning.items():
            atom_plane_points = partitioning_df["plane_points"].to_list()
            atom_plane_vectors = partitioning_df["plane_vectors"].to_list()
            plane_points.extend(atom_plane_points)
            plane_vectors.extend(atom_plane_vectors)

        # convert plane points and vectors to arrays and then convert to the plane
        # equation
        plane_points = np.array(plane_points)
        plane_vectors = np.array(plane_vectors)
        return plane_points, plane_vectors

    @cached_property
    def all_plane_equations(self):
        """
        A (N,4) array containing every partitioning plane equation
        """
        plane_points, plane_vectors = self.all_partitioning_plane_points_and_vectors
        D = -np.sum(plane_points * plane_vectors, axis=1)
        # convert to all coefficients
        return np.column_stack((plane_vectors, D))

    @cached_property
    def number_of_planes_per_atom(self):
        """
        A list for splitting an array containing all of the partitioning planes
        back into atom based sections.
        """
        partitioning = self.partitioning
        number_of_planes_per_atom = [len(planes) for planes in partitioning.values()]
        number_of_planes_per_atom.pop(-1)
        number_of_planes_per_atom = np.array(number_of_planes_per_atom)
        sum_number_of_planes_per_atom = []
        for i, j in enumerate(number_of_planes_per_atom):
            sum_number_of_planes_per_atom.append(
                j + np.sum(number_of_planes_per_atom[:i])
            )
        return sum_number_of_planes_per_atom

    @cached_property
    def voxel_edge_vectors(self):
        """
        A (12,3) array consisting of the vectors that make up the edges of
        a voxel in the grid
        """
        # Here we define the edges of each voxel in terms of vertex indices. these are
        # not unique and other choices could be made. One potentially faster option is
        # to arrange the edges so that the first index is as small of a set of indices
        # as possible. Later when we calculate t in our line segments this would reduce
        # the number of calculations needed for the numerator which only depends on
        # the plane equation and the first vertex position
        # Get the edge vectors in cartesian coordinates
        vertices_transforms_cart = self.vertices_transforms_cart
        edges = np.array(
            [
                [0, 1],
                [0, 2],
                [0, 4],
                [3, 1],
                [3, 2],
                [3, 7],
                [5, 1],
                [5, 4],
                [5, 7],
                [6, 2],
                [6, 4],
                [6, 7],
            ]
        )
        edge_vectors = []
        for edge in edges:
            edge_vector = (
                vertices_transforms_cart[edge[1]] - vertices_transforms_cart[edge[0]]
            )
            edge_vectors.append(edge_vector)
        edge_vectors = np.array(edge_vectors)
        return edge_vectors

    def get_site_assignments_from_frac_coords(
        self,
        voxel_frac_coords: ArrayLike,
        min_dist_from_plane: float,
    ):
        """
        Gets the site assignments for an arbitrary number of voxels described
        by their fractional coordinates.

        Args:
            voxel_frac_coords (ArrayLike):
                An N,3 array of fractional coordinates corresponding to the voxels
                to assign sites to
            min_dist_from_plane (float):
                The minimum distance a point should be from the plane before
                a site can be assigned. This is value usually corresponds to
                the maximum distance the voxel center can be from the plane
                while the voxel is still being intersected by the plane.

        Returns:
            A 1D array of atomic site assignments. Assignments start at 1 with
            0 indicating no site was found for this coordinate.
        """
        grid = self.charge_grid
        plane_equations = self.all_plane_equations
        number_of_planes_per_atom = self.number_of_planes_per_atom
        unit_cell_permutations_frac = self.unit_cell_permutations_frac

        # Create an array of zeros to map back to
        zeros_array = np.zeros(len(voxel_frac_coords))
        # Create an array that the results will be added to
        results_array = zeros_array.copy()

        # create zeros array for any problems
        # global_indices_to_zero = np.array([])
        # check every possible permutation
        for transformation in tqdm(
            unit_cell_permutations_frac,
            total=len(unit_cell_permutations_frac),
            ascii="░▒▓",
        ):
            # Get the indices where voxels haven't been assigned. Get only these
            # frac coords
            indices_where_zero = np.where(results_array == 0)[0]
            new_frac_coords = voxel_frac_coords.copy()[indices_where_zero]
            # transform the fractional coords to the next transformation
            x1, y1, z1 = transformation
            new_frac_coords[:, 0] += x1
            new_frac_coords[:, 1] += y1
            new_frac_coords[:, 2] += z1
            # Convert the frac coords into cartesian coords
            cart_coords = grid.get_cart_coords_from_frac_full_array(
                new_frac_coords
            ).astype(float)
            points = np.array(cart_coords).astype(float)
            planes = np.array(plane_equations).astype(float)
            # There is a difference in the speed of dask vs numpy. Dask has a
            # lot of overhead, but at a certain point it is faster than numpy.
            # We check which one we should use here.
            plane_distances_to_calc = len(points) * len(planes)
            if plane_distances_to_calc > 7.8e8:
                dask = True
            else:
                dask = False

            if dask:
                # DASK ARRAY VERSION
                # points = da.from_array(points)
                # planes = da.from_array(planes)
                distances = da.dot(points, planes[:, :3].T) + planes[:, 3]
                # Round the distances to within 5 decimals. Everything to this point has
                # been based on lattice position from vasp which typically have 5-6
                # decimal places (7 sig figs)
                distances = np.round(distances, 12)
                # We write over the distances with a more simplified boolean to save
                # space. This is also where we filter if we're near a plane if desired
                distances = da.where(distances < -min_dist_from_plane, True, False)
                distances = distances.compute()

            else:
                # BASE NUMPY VERSION
                distances = np.dot(points, planes[:, :3].T) + planes[:, 3]
                # Round the distances to within 5 decimals. Everything to this point has
                # been based on lattice position from vasp which typically have 5-6
                # decimal places (7 sig figs)
                distances = np.round(distances, 12)
                # We write over the distances with a more simplified boolean to save
                # space. This is also where we filter if we're near a plane if desired
                distances = np.where(distances < -min_dist_from_plane, True, False)

            # split the array into the planes belonging to each atom. Again we write
            # over to save space
            distances = np.array_split(distances, number_of_planes_per_atom, axis=1)
            # get a 1D array representing the voxel indices with the atom index where the
            # voxel is assigned to a site and 0s where they are not
            new_results_arrays = []
            # for atom_index, atom_array in enumerate(distances_split_by_atom):
            for atom_index, atom_array in enumerate(distances):
                voxel_result = np.all(atom_array, axis=1)
                voxel_result = np.where(
                    voxel_result == True, atom_index + 1, voxel_result
                )
                new_results_arrays.append(voxel_result)

            indices_to_zero = []
            new_results_array = np.zeros(len(new_results_arrays[0]))
            for i, sub_results_array in enumerate(new_results_arrays):
                new_results_array = np.sum(
                    [new_results_array, sub_results_array], axis=0
                )
                indices_to_zero.extend(np.where(new_results_array > i + 1)[0])
            indices_to_zero = np.unique(indices_to_zero).astype(int)
            # print(len(indices_to_zero))
            new_results_array[indices_to_zero] = 0
            # Sum the results
            # new_results_array = np.sum(new_results_arrays, axis=0)
            # add results to the results_array
            results_array[indices_where_zero] = new_results_array
        return results_array

    def get_site_assignments_from_frac_coords_with_memory_handling(
        self,
        voxel_frac_coords: ArrayLike,
        min_dist_from_plane: float,
    ):
        """
        Gets the site assignments for an arbitrary number of voxels described
        by their fractional coordinates. Takes available memory into account
        and divides the voxels into chunks to perform operations.

        Args:
            voxel_frac_coords (ArrayLike):
                An N,3 array of fractional coordinates corresponding to the voxels
                to assign sites to
            min_dist_from_plane (float):
                The minimum distance a point should be from the plane before
                a site can be assigned. This is value usually corresponds to
                the maximum distance the voxel center can be from the plane
                while the voxel is still being intersected by the plane.

        Returns:
            A 1D array of atomic site assignments. Assignments start at 1 with
            0 indicating no site was found for this coordinate.
        """
        partitioning = self.partitioning
        # determine how much memory is available. Then calculate how many distance
        # calculations would be possible to do at once with this much memory.
        available_memory = psutil.virtual_memory().available / (1024**2)

        handleable_plane_distance_calcs_numpy = available_memory / 0.00007
        handleable_plane_distance_calcs_dask = available_memory / 0.000025
        plane_distances_to_calc = len(voxel_frac_coords) * sum(
            [len(i) for i in partitioning.values()]
        )

        # I found there is a cutoff where Dask becomes faster than numpy. This
        # may vary with the number of cores available. It is largely due to Dask
        # having a large overhead.
        if plane_distances_to_calc > 7.8e8:
            # calculate the number of chunks the voxel array should be split into to not
            # overload the memory. Then split the array by this number
            split_num = math.ceil(
                plane_distances_to_calc / handleable_plane_distance_calcs_dask
            )
        else:
            split_num = math.ceil(
                plane_distances_to_calc / handleable_plane_distance_calcs_numpy
            )
        split_voxel_frac_coords = np.array_split(voxel_frac_coords, split_num, axis=0)
        # create an array to store results
        voxel_results_array = np.array([])
        # for each split, calculate the results and add to the end of our results
        for chunk, split_voxel_array in enumerate(split_voxel_frac_coords):
            logging.info(
                f"Calculating site assignments for voxel chunk {chunk+1}/{split_num}"
            )
            split_result = self.get_site_assignments_from_frac_coords(
                voxel_frac_coords=split_voxel_array,
                min_dist_from_plane=min_dist_from_plane,
            )
            voxel_results_array = np.concatenate([voxel_results_array, split_result])
        return voxel_results_array

    def get_single_site_voxel_assignments(self, all_site_voxel_assignments: ArrayLike):
        """
        Gets the voxel assignments for voxels that are not split by a plane.

        Args:
            all_site_voxel_assignments (ArrayLike):
                A 1D array of integers representing the site assignments for
                each voxel in the grid.

        Returns:
            A 1D array of the same length as the input with additional site
            assignments.
        """
        all_voxel_assignments = all_site_voxel_assignments.copy()
        # charge_grid = self.charge_grid
        # In the BadELF algorithm the electride sites will have already been
        # assigned. In VoronELF they won't be. Here we search for unassigned
        # voxels and then run the alg on the remaining ones
        unassigned_indices = np.where(all_voxel_assignments == 0)[0]
        all_voxel_frac_coords = self.all_voxel_frac_coords
        frac_coords_to_find = all_voxel_frac_coords[unassigned_indices]
        # min_dist_from_plane = charge_grid.max_voxel_dist
        single_site_voxel_assignments = (
            self.get_site_assignments_from_frac_coords_with_memory_handling(
                frac_coords_to_find, min_dist_from_plane=0
            )
        )
        all_voxel_assignments[unassigned_indices] = single_site_voxel_assignments
        return all_voxel_assignments

    def get_plane_split_voxel_assignments(self, voxel_frac_coords):
        """
        Gets the voxel assignments for voxels that are exactly on a plane.

        Args:
            voxel_frac_coords (ArrayLike):
                A 2D array representing the fractional coordinates of voxels
                that haven't been assigned

        Returns:
            A 2D array of shape (N,M) where N is the length of the input array
            and M is the number of atoms. A 1 means the atom is assigned this
            voxel.
        """
        logging.info("Calculating assignments for voxels on planes")
        if self.algorithm == "badelf":
            structure = self.charge_grid.structure
        elif self.algorithm == "voronelf":
            structure = self.electride_structure
        grid = self.charge_grid
        plane_equations = self.all_plane_equations
        number_of_planes_per_atom = self.number_of_planes_per_atom
        unit_cell_permutations_frac = self.unit_cell_permutations_frac

        # Create an array of zeros to map back to
        zeros_array = np.zeros([len(voxel_frac_coords), len(structure)])
        # Create an array that the results will be added to
        results_array = zeros_array.copy()

        # create zeros array for any problems
        # global_indices_to_zero = np.array([])
        # check every possible permutation
        for transformation in tqdm(
            unit_cell_permutations_frac,
            total=len(unit_cell_permutations_frac),
            ascii="░▒▓",
        ):
            # Get the indices where voxels haven't been assigned. Get only these
            # frac coords
            new_frac_coords = voxel_frac_coords.copy()
            # transform the fractional coords to the next transformation
            x1, y1, z1 = transformation
            new_frac_coords[:, 0] += x1
            new_frac_coords[:, 1] += y1
            new_frac_coords[:, 2] += z1
            # Convert the frac coords into cartesian coords
            cart_coords = grid.get_cart_coords_from_frac_full_array(
                new_frac_coords
            ).astype(float)
            points = np.array(cart_coords).astype(float)
            planes = np.array(plane_equations).astype(float)
            # There is a difference in the speed of dask vs numpy. Dask has a
            # lot of overhead, but at a certain point it is faster than numpy.
            # We check which one we should use here.
            plane_distances_to_calc = len(points) * len(planes)
            if plane_distances_to_calc > 7.8e8:
                dask = True
            else:
                dask = False

            if dask:
                # DASK ARRAY VERSION
                # points = da.from_array(points)
                # planes = da.from_array(planes)
                distances = da.dot(points, planes[:, :3].T) + planes[:, 3]
                # Round the distances to within 5 decimals. Everything to this point has
                # been based on lattice position from vasp which typically have 5-6
                # decimal places (7 sig figs)
                distances = np.round(distances, 12)
                # We write over the distances with a more simplified boolean to save
                # space. This is also where we filter if we're near a plane if desired
                distances = da.where(distances <= 0, True, False)
                distances = distances.compute()

            else:
                # BASE NUMPY VERSION
                distances = np.dot(points, planes[:, :3].T) + planes[:, 3]
                # Round the distances to within 5 decimals. Everything to this point has
                # been based on lattice position from vasp which typically have 5-6
                # decimal places (7 sig figs)
                distances = np.round(distances, 12)
                # We write over the distances with a more simplified boolean to save
                # space. This is also where we filter if we're near a plane if desired
                distances = np.where(distances <= 0, True, False)

            # split the array into the planes belonging to each atom. Again we write
            # over to save space
            distances = np.array_split(distances, number_of_planes_per_atom, axis=1)
            # get a 1D array representing the voxel indices with the atom index where the
            # voxel is assigned to a site and 0s where they are not
            new_results_arrays = []
            # for atom_index, atom_array in enumerate(distances_split_by_atom):
            for atom_index, atom_array in enumerate(distances):
                voxel_result = np.all(atom_array, axis=1)
                voxel_result = np.where(
                    voxel_result == True, atom_index + 1, voxel_result
                )
                new_results_arrays.append(voxel_result)

            # add results to the results_array
            results_array += np.column_stack(new_results_arrays)

        results_array = np.where(results_array > 0, 1, 0)
        return results_array

    def get_multi_site_voxel_assignments(self, all_site_voxel_assignments):
        """
        Gets the voxel assignments for voxels that are not split by a plane.

        Args:
            all_site_voxel_assignments (ArrayLike):
                A 1D array of integers representing the site assignments for
                each voxel in the grid.

        Returns:
            A 2D array of fractional site assignments with indices (i,j) where i is the
            voxel index and j is the site index.
        """
        grid = self.charge_grid
        # Get all site assignments so far
        all_site_voxel_assignments = all_site_voxel_assignments.copy()
        # Check that there are any unassigned voxels. If none, return an empty
        # array
        if len(np.where(all_site_voxel_assignments == 0)[0]) == 0:
            return np.array([])
        # The following lines are a more rigorous method of calculating the voxels
        # directly on planes. It takes longer so it's commented for now.
        # Now we want to get assignments for voxels directly on planes
        # all_frac_coords = grid.all_voxel_frac_coords
        # unassigned_indices_1d = np.where(all_site_voxel_assignments==0)[0]
        # unassigned_frac_coords = all_frac_coords[unassigned_indices_1d]
        # on_plane_assignment = self.get_plane_split_voxel_assignments(unassigned_frac_coords)
        # Some of the voxels will have been assigned. We want to store the indices of
        # these so we can zero them out later

        # Get as unassigned grid
        assignment_array = all_site_voxel_assignments.reshape(grid.shape).astype(
            np.int16
        )
        unassigned_indices = np.column_stack(np.where(assignment_array == 0))
        # If doing self we don't want assignments to go to electrides so we zero them
        # here. We need to save them to put them back later though
        if self.algorithm == "badelf":
            electride_indices = (
                np.array(self.electride_structure.indices_from_symbol("He")) + 1
            )
            assignment_array = np.where(
                np.isin(assignment_array, electride_indices), 0, assignment_array
            )
            structure = grid.structure
        elif self.algorithm == "voronelf":
            structure = self.electride_structure
        # Get smallest radius for a sphere
        lattice = grid.structure.lattice

        # Get the starting radius/min_radius and the maximum radius we could possibly
        # go to
        nearest_voxels = lattice.abc / grid.shape
        min_radius = nearest_voxels.max()
        test_radius = min_radius.copy()
        max_radius = (np.array(lattice.abc) / 2).max()

        # Get array to assign to
        # new_assignment_array = on_plane_assignment.copy()
        new_assignment_array = np.zeros([len(unassigned_indices), len(structure)])

        while len(np.where(np.sum(new_assignment_array, axis=1) == 0)[0]) > 0:
            if test_radius >= max_radius:
                raise Exception()
            # Get the transformations to get a sphere around a voxel
            sphere_transforms = grid.get_voxels_transformations_to_radius(test_radius)
            # Get the indices that still don't have an assignment
            still_unassigned_1d_indices = np.where(
                np.sum(new_assignment_array, axis=1) == 0
            )[0]
            # Get the voxel coords of the unassigned indices
            unassigned_voxel_coords = unassigned_indices[still_unassigned_1d_indices]
            possible_assignments = []
            # for each voxel in the sphere around the voxel, get the site assignment
            for transform in sphere_transforms:
                new_indices = unassigned_voxel_coords + transform
                # wrap around indices
                new_x = (new_indices[:, 0] % grid.shape[0]).astype(int)
                new_y = (new_indices[:, 1] % grid.shape[1]).astype(int)
                new_z = (new_indices[:, 2] % grid.shape[2]).astype(int)
                site_assignment = assignment_array[new_x, new_y, new_z]
                possible_assignments.append(site_assignment)
            possible_assignments = np.column_stack(possible_assignments)
            # Get the assignments that are not still 0
            non_zero_indices = np.where(np.sum(possible_assignments, axis=1) > 0)[0]
            sub_possible_assignments = possible_assignments[non_zero_indices]
            # iterate over these assignments and count how many there are of each.
            # Update the corresponding row in the new assignments array with total
            # count for each site.
            for i, possible_label in enumerate(sub_possible_assignments):
                sites, counts = np.unique(possible_label, return_counts=True)
                for site, count in zip(sites[1:], counts[1:]):
                    new_assignment_array[
                        still_unassigned_1d_indices[non_zero_indices[i]], site - 1
                    ] = count
            # increase our test radius for the next round
            test_radius += min_radius
            # print(test_radius, len(np.where(np.sum(new_assignment_array,axis=1)==0)[0]))

        # We need to rescale the assignment array to check for any that have assignments
        # much higher than 0.5
        new_assignment_array_sum = np.tile(
            np.sum(new_assignment_array, axis=1), (len(new_assignment_array[0]), 1)
        ).T
        new_assignment_array = new_assignment_array / new_assignment_array_sum
        return new_assignment_array



# -*- coding: utf-8 -*-

import logging
import math
import warnings
from functools import cached_property
from itertools import combinations

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from pybader.interface import Bader
from pymatgen.analysis.local_env import CrystalNN
from scipy.interpolate import RegularGridInterpolator
from scipy.signal import savgol_filter
from scipy.spatial import ConvexHull
from tqdm import tqdm

from simmate.apps.badelf.core.grid import Grid
from simmate.toolkit import Structure


class PartitioningToolkit:
    """
    A set of tools for aiding in partitioning a 3D voxel grid. These tools make
    up the basis for the partitioning utilized in BadELF and the related
    VoronELF algorithms.

    Args:
        grid (Grid):
            A BadELF app Grid type object.
        pybader (Bader):
            A bader Bader type object.
    """

    def __init__(self, grid: Grid, bader: Bader):
        self.grid = grid.copy()
        self.bader = bader

    def get_partitioning_line_from_voxels(
        self,
        site_voxel_coord: ArrayLike | list,
        neigh_voxel_coord: ArrayLike | list,
        method: str = "linear",
        steps: int = 200,  #!!! This should be set dynamically in the future
    ):
        """
        Finds a line of voxel positions between two atom sites and then finds the value
        of the partitioning grid at each of these positions. The values are found
        using an interpolation function defined using SciPy's RegularGridInterpeter.

        Args:
            site_voxel_coord (ArrayLike):
                The voxel coordinates of an atomic site
            neigh_voxel_coord (ArrayLike):
                The voxel coordinates of a neighboring
                site
            method (str):
                The method of interpolation. 'cubic' is more rigorous
                than 'linear'
            steps (int):
                The number of voxel coordinates to interpolate. Default is 200

        Results:
            A list with 200 pairs of voxel coordinates and data values along
            a line between two positions.
        """
        grid_data = self.grid.copy().total
        label_data = self.bader.atoms_volumes
        slope = [b - a for a, b in zip(site_voxel_coord, neigh_voxel_coord)]
        slope_increment = [float(x) / steps for x in slope]

        # get a list of points along the connecting line. First add the original
        # site
        position = site_voxel_coord
        line = [[round(float(a % b), 12) for a, b in zip(position, grid_data.shape)]]
        for i in range(steps):
            # move position by slope_increment
            position = [float(a + b) for a, b in zip(position, slope_increment)]

            # Wrap values back into cell
            # We must do (a-1) to shift the voxel index (1 to grid_max+1) onto a
            # normal grid, (0 to grid_max), then do the wrapping function (%), then
            # shift back onto the VASP voxel index.
            position = [
                round(float(a % b), 12) for a, b in zip(position, grid_data.shape)
            ]

            line.append(position)

        # The partitioning uses a padded grid and grid interpolation to find the
        # location of dividing planes.
        padded_grid_data = np.pad(grid_data, 1, mode="wrap")
        padded_label_data = np.pad(label_data, 1, mode="wrap")

        # interpolate grid to find values that lie between voxels. This is done
        # with a cruder interpolation here and then the area close to the minimum
        # is examened more closely with a more rigorous interpolation in
        # get_line_frac_min
        a, b, c = self.grid.get_padded_grid_axes(1)
        fn = RegularGridInterpolator((a, b, c), padded_grid_data, method=method)
        fn_label = RegularGridInterpolator((a, b, c), padded_label_data, "nearest")
        # get a list of the ELF values along the line
        values = []
        label_values = []

        for pos in line:
            adjusted_pos = [x + 1 for x in pos]
            value = float(fn(adjusted_pos))
            label_value = int(fn_label(adjusted_pos))
            values.append(value)
            label_values.append(label_value)

        return line, values, label_values

    def get_partitioning_line_from_indices(
        self, i: int, j: int, method: str = "linear"
    ):
        """
        Gets the voxel positions and elf values for points between two sites in
        the structure.

        Args:
            i (int):
                index of first site in the structure
            j (int):
                index of second site in the structure
            method (str):
                The method of interpolation. 'cubic' is more rigorous
                than 'linear'

        Returns:
        - Two lists, one of positions in voxel coordinates and another of elf
        values
        """
        grid = self.grid.copy()
        site_voxel_coord = grid.get_voxel_coords_from_index(i)
        neigh_voxel_coord = grid.get_voxel_coords_from_index(j)
        return self.get_partitioning_line_from_voxels(
            site_voxel_coord, neigh_voxel_coord, method=method
        )

    def get_partitioning_line_from_cart_coords(
        self,
        site_cart_coords: ArrayLike | list,
        neigh_cart_coords: ArrayLike | list,
        method: str = "linear",
    ):
        """
        Gets the voxel positions and elf values for points between two sites in
        the structure given as cartesian coordinates. This method can also be
        used to find the values in the ELF between two arbitrary points in the
        structure.

        Args:
            site_cart_coords (ArrayLike):
                cartesian coordinates of a site in the structure
            neigh_cart_coords (ArrayLike):
                cartesian coordinates of a second site in the structure
            method (str):
                The method of interpolation. 'cubic' is more rigorous
                than 'linear'

        Returns:
            Two lists, one of positions in voxel coordinates and another of elf
            values
        """
        grid = self.grid.copy()
        site_voxel_coord = grid.get_voxel_coords_from_cart(site_cart_coords)
        neigh_voxel_coord = grid.get_voxel_coords_from_cart(neigh_cart_coords)
        return self.get_partitioning_line_from_voxels(
            site_voxel_coord, neigh_voxel_coord, method=method
        )

    @staticmethod
    def _check_partitioning_line_for_symmetry(values: list, tolerance: float = 10):
        """
        Check if the values are roughly symmetric. Checks each value versus
        the equivalent value of the other half to see if they are within the
        requested percent diff.

        Args:
            values (list):
                List of numeric values
            tolerance (float):
                Tolerance level for symmetry check in percents.

        Returns:
            True if roughly symmetric, False otherwise
        """
        n = len(values)

        # Check if the list has an even number of elements
        if n % 2 != 0:
            # remove the center if odd number
            center_index = math.ceil(n / 2)
            values.pop(center_index)

        # Split the list into two halves
        half_size = n // 2
        first_half = values[:half_size]
        second_half = values[half_size:]

        # Reverse the second half
        reversed_second_half = list(reversed(second_half))

        # Check if the values are roughly equal within the given tolerance
        for val1, val2 in zip(first_half, reversed_second_half):
            perc_diff = abs((val1 - val2) / (val1 + val2)) * 100
            if perc_diff > tolerance:
                return False

        return True

    @staticmethod
    def find_minimum(values: list | ArrayLike):
        """
        Finds the local minima in a list of values and returns the index and value
        at each as a list of form [[min_index1, min_value1], [min_index2, min_value2], ...]

        Args:
            values (list):
                The list of values to find the minima of

        results:
            A list of minima represented by [index, value]
        """
        minima = [
            [i, y]
            for i, y in enumerate(values)
            if ((i == 0) or (values[i - 1] >= y))
            and ((i == len(values) - 1) or (y < values[i + 1]))
        ]
        return minima

    @staticmethod
    def find_maximum(values: list | ArrayLike):
        """
        Finds the local maxima in a list of values and returns the index and value
        at each as a list of form [[max_index1, max_value1], [max_index2, max_value2], ...]

        Args:
            values (list):
                The list of values to find the minima of

        results:
            A list of maxima represented by [index, value]
        """
        maxima = [
            [i, y]
            for i, y in enumerate(values)
            if ((i == 0) or (values[i - 1] <= y))
            and ((i == len(values) - 1) or (y > values[i + 1]))
        ]
        return maxima

    @staticmethod
    def get_closest_extrema_to_center(
        values: list | ArrayLike,
        extrema: list | ArrayLike,
    ):
        """
        Takes a list of values and the relative extrema (either minima or maxima)
        and finds which extrema is closest to the center of the line.

        Args:
            values (list):
                A list of values
            extrema (list):
                A list of extrema of form [index, value]

        results:
            The global extreme of form [index, value]
        """
        midpoint = len(values) / 2
        differences = []
        for pos, val in extrema:
            diff = abs(pos - midpoint)
            differences.append(diff)
        min_pos = differences.index(min(differences))
        global_extrema = extrema[min_pos]
        return global_extrema

    def get_line_minimum_as_frac(
        self,
        positions: list,
        values: list | ArrayLike,
        labels: list | ArrayLike,
        site_index: int,
        neigh_index: int,
    ):
        """
        Finds the minimum point of a list of values along a line, then returns the
        fractional position of this values position along the line.

        Args:
            positions (list):
                A list of positions given as voxel coordinates along the line
                of interest
            values (list):
                A list of values to find the minimum of
            site_index (int):
                The index of the atom
            neigh_index (int):
                The index of the neighboring atom

        results:
            The global minimum of form [line_position, value, frac_position]
        """
        # In some cases I have found that linear interpolation near electride
        # sites is wavy resulting in many maxima and causing problems with the
        # partitioning. To deal with this I put the linear values through a
        # savgol filter
        values = savgol_filter(values, 20, 3)

        site_equiv = self.grid.equivalent_atoms[site_index]
        neigh_equiv = self.grid.equivalent_atoms[neigh_index]
        # get the string for the site and neigh. During the electride dimensionality
        # search this can throw an error so we add a try/except clause here.
        try:
            site_string = self.grid.structure[site_equiv].species_string
            neigh_string = self.grid.structure[neigh_equiv].species_string
        except:
            site_string = "He"
            neigh_string = "H"

        if site_equiv == neigh_equiv:
            # We have an atom bonded to another atom that is symmetrically equivalent
            # so the resulting elf line will be equivalent
            symmetric = True
        elif site_string == neigh_string:
            list_values = list(values)
            # We have two atoms of the same type that are not found as symmetric
            # by pymatgen. However, it is still likely that the ELF between the
            # two atoms is symmetric which can easily result in a covalent like
            # behavior. To handle this, we check if the line is symmetric and
            # if it is we place the frac at exactly the middle
            symmetric = self._check_partitioning_line_for_symmetry(list_values)
        else:
            symmetric = False

        # Now if our line is symmetric, we place our partitioning exactly at
        # the middle of the line. Otherwise we interpolate to find where to
        # place the line.
        if symmetric:
            elf_value = values[100]
            elf_min_frac = 0.5
            global_min = [100, elf_value, elf_min_frac]
        else:
            # If the only assignments in our line are for the two atoms involved
            # we guess that our initial partitioning frac will be at the furthest
            # point still labeled as belonging to the site
            if np.all(np.isin(labels, [site_index, neigh_index])):
                elf_min_index = np.where(np.array(labels) == site_index)[0].max()
                extrema = "min"
            else:
                # There is at least some section of the line that is assigned to an
                # atom not in the bond. We want to assign the fraction as being at
                # the maximum of this unrelated area. However, sometimes I've
                # found that there is no maximum in the range of this area which
                # causes the assignment to be placed very far from what is reasonable.
                # to handle this we check if there is a maximum in the unrelated area
                # and if not we assign as above.
                # There is a potential for a bug where a very small maximum is
                # found using linear interpolation that is then removed with
                # cubic interpolation. This actually happened with Ba2N. To deal
                # with this, I've added a cutoff where the maximum must be at
                # least 0.01 greater than the absolute minimum of the values
                maxima = self.find_maximum(values)
                unrelated_indices = np.where(
                    ~np.isin(labels, [site_index, neigh_index])
                )
                new_maxima = []
                for maximum in maxima:
                    if (
                        np.isin(maximum[0], unrelated_indices)
                        and (maximum[1] - min(values)) > 0.01
                    ):
                        new_maxima.append(maximum)
                if len(new_maxima) > 0:
                    elf_min_index = self.get_closest_extrema_to_center(
                        values, new_maxima
                    )[0]
                    extrema = "max"
                else:
                    elf_min_index = np.where(np.array(labels) == site_index)[0].max()
                    extrema = "min"

            global_min = self._refine_line_part_frac(
                positions=positions,
                elf_min_index=elf_min_index,
                extrema=extrema,
            )

        return global_min

    def _refine_line_part_frac(
        self,
        positions: list,
        elf_min_index: int,
        extrema: str,
    ):
        """
        Refines the location of the minimum along an ELF line between two sites.
        To do this, the initial estimate from a linear interpolation of the line
        is used and a cubic interpolation is used in a smaller area around the
        estimated point. The sampled area is adjusted if it is found to not be
        centered on the new more accurate minimum.

        Args:
            positions (list):
                A list of positions given as voxel coordinates along the line
                of interest
            elf_min_index (int):
                The index along the line at which the linear interpolation estimated
                the minimum.
            extrema (str):
                Which type of extrema to refine. Either max or min.

        Returns:
            The global minimum of form [line_position, value, frac_position]
        """
        amount_to_pad = 10
        grid = self.grid.copy()
        padded = np.pad(grid.total, amount_to_pad, mode="wrap")

        # interpolate the grid with a more rigorous method to find more exact value
        # for the plane.
        a, b, c = grid.get_padded_grid_axes(10)
        fn = RegularGridInterpolator((a, b, c), padded, method="cubic")

        # create variables for if the line needs to be shifted from what the
        # rough partitioning found
        centered = False
        amount_to_shift = 0
        attempts = 0

        while centered == False:
            if attempts == 5:
                break
            else:
                attempts += 1
                # If the position wasn't centered previously, we need to shift
                # the index
                elf_min_index = elf_min_index + amount_to_shift
                line_section = positions[elf_min_index - 3 : elf_min_index + 4]
                line_section_x = [
                    i for i in range(elf_min_index - 3, elf_min_index + 4)
                ]

                values_fine = []
                # Get the list of values from the interpolated grid
                for pos in line_section:
                    new_pos = [i + amount_to_pad for i in pos]
                    value_fine = float(fn(new_pos))
                    values_fine.append(value_fine)

                # Find the minimum value of this line as well as the index for this value's
                # position.
                try:
                    if extrema == "min":
                        minimum_value = min(values_fine)
                    elif extrema == "max":
                        minimum_value = max(values_fine)
                except:
                    attempts = 5
                    continue
                min_pos = values_fine.index(minimum_value)  # + global_min_pos[0]-5

                if min_pos == 4:
                    # Our line is centered and we can move on
                    centered = True
                else:
                    # Our line is not centered and we need to adjust it
                    amount_to_shift = min_pos - 4

        if not centered:
            # The above sometimes fails because the linear fitting gives a guess
            # for the minimum that isn't close. To handle this we treat these
            # situations rigorously
            values = []

            # Get the ELF value for every position in the line.
            for pos in positions:
                new_pos = [i + amount_to_pad for i in pos]
                value = float(fn(new_pos))
                values.append(value)

            # Get a list of all of the minima along the line
            if extrema == "min":
                minima = self.find_minimum(values)
            elif extrema == "max":
                minima = self.find_maximum(values)

            # then we grab the local minima closest to the midpoint of the line
            global_min = self.get_closest_extrema_to_center(values, minima)

            # now we want a small section of the line surrounding the minimum
            values_fine = values[global_min[0] - 3 : global_min[0] + 4]
            line_section_x = [i for i in range(global_min[0] - 3, global_min[0] + 4)]

        # now that we've found the values surrounding the minimum of our line,
        # we can fit these values to a 2nd degree polynomial and solve for its
        # minimum point
        try:
            d, e, f = np.polyfit(line_section_x, values_fine, 2)
            x = -e / (2 * d)
            elf_min_index_new = x
            elf_min_value_new = np.polyval(np.array([d, e, f]), x)
            elf_min_frac_new = elf_min_index_new / (len(positions) - 1)
        except:
            breakpoint()

        return [elf_min_index_new, elf_min_value_new, elf_min_frac_new]

    @staticmethod
    def get_plane_sign(
        point: ArrayLike | list,
        plane_vector: ArrayLike | list,
        plane_point: ArrayLike | list,
    ):
        """
        Gets the sign associated with a point compared with a plane.

        Args:
            point (ArrayLike):
                A point in cartesian coordinates to compare with a plane
            plane_vector (ArrayLike):
                The vector normal to the plane of interest
            plane_point (ArrayLike):
                A point on the plane of interest

        Returns:
            The sign of the point compared with the plane and the distance of
            the point to the plane.
        """
        # get all of the points in cartesian coordinates
        x, y, z = plane_point
        a, b, c = plane_vector
        x1, y1, z1 = point
        value_of_plane_equation = round(
            (a * (x - x1) + b * (y - y1) + c * (z - z1)), 12
        )
        # get the distance of the point from the plane with some allowance of error.
        if value_of_plane_equation > 0:
            return "positive", value_of_plane_equation
        elif value_of_plane_equation < 0:
            return "negative", value_of_plane_equation
        else:
            return "zero", value_of_plane_equation

    def get_elf_ionic_radius(
        self,
        site_index: int,
    ):
        """
        This method gets the ELF ionic radius. It interpolates the ELF values
        between a site and it's closest neighbor and returns the distance between
        the atom and the minimum in this line. This has been shown to be very
        similar to the Shannon Crystal Radius, but gives more specific values

        Args:
            site_index (int):
                An integer value referencing an atom in the structure

        Returns:
            The distance the ELF ionic radius of the site
        """
        # get closest neighbor for the given site

        neighbors = self.all_site_neighbor_pairs
        # get only this sites dataframe
        site_df = neighbors.loc[neighbors["site_index"] == site_index]
        site_df.reset_index(inplace=True, drop=True)

        # Get to the closest neighbor to the site that isn't a He dummy atom
        for i, row in site_df.iterrows():
            site_cart_coords = row["site_coords"]
            neigh_cart_coords = row["neigh_coords"]
            neighbor_string = row["neigh_symbol"]
            if neighbor_string != "He":
                bond_dist = row["dist"]
                break

        (
            elf_positions,
            elf_values,
            label_values,
        ) = self.get_partitioning_line_from_cart_coords(
            site_cart_coords,
            neigh_cart_coords,
        )

        elf_min_index = np.where(np.array(label_values) == site_index)[0].max()
        refined_index = self._refine_line_part_frac(elf_positions, elf_min_index, "min")
        distance_to_min = refined_index[2] * bond_dist

        return distance_to_min

    @staticmethod
    def _get_vector_plane_intersection(
        point0: ArrayLike | list,
        point1: ArrayLike | list,
        plane_point: ArrayLike | list,
        plane_vector: ArrayLike | list,
        allow_point_intercept: bool = False,
    ):
        """
        Takes in two points and the point/vector defining a plane and returns
        the point where the line segment and plane intersect (if it exists)

        Args:
            point0 (ArrayLike): The first point of a line segment
            point1 (ArrayLike): The second point of a line segment
            plane_point (ArrayLike): A point on the plane
            plane_vector (ArrayLike): The vector normal to the plane
            allow_point_intercept (bool): Whether to count a point at the end
                of the line segment touching the plane as an intercept.

        Returns:
            The point where the line segment intersects the plane or None if
            there is no intersection.
        """
        # convert points to NumPy arrays
        point0 = np.array(point0)
        point1 = np.array(point1)
        plane_point = np.array(plane_point)
        plane_vector = np.array(plane_vector)

        # get direction of line segment.
        direction = point1 - point0
        # get dot product of direction vector and dot_product
        dot_product = np.dot(direction, plane_vector)
        # check if line is parallel to plane
        if np.abs(dot_product) < 1e-06:
            return None

        # get distance from the line segment point to the plane
        distance = np.dot(plane_point - point0, plane_vector) / dot_product

        # calculate intersection point
        intersection_point = point0 + direction * distance
        # round the intersection points

        # check if intersection point is between the start and end points of our
        # line segment. To do this, we would normally first check if the point
        # is on the line, but it must be alredy because we defined it as such. Next
        # we check the dot products
        # Get the vector for the intersecting point and point1 assuming point0 is
        # the origin
        AB = point1 - point0
        AC = intersection_point - point0
        # Get the dot products of the intersecting point and point1
        KAC = np.dot(AB, AC)
        KAB = np.dot(AB, AB)
        # There are five possible scenarios.
        if KAC < -1e-8 or KAC > KAB + 1e-8:
            # the point is outside the line
            return None
        elif KAC == 0 or KAC == KAB:
            # The point is at point0 or point1
            if allow_point_intercept:
                return intersection_point
            else:
                return None
        else:
            # The point is between point0 and point1
            return intersection_point

    def _reduce_planes_quick(
        self,
        plane_points: ArrayLike | list,
        plane_vectors: ArrayLike | list,
        dists: ArrayLike | list,
        site_cart_coord: ArrayLike | list,
    ):
        """
        Reduces a set of planes around a site by checking if the line between
        the site and each plane is intersected by other planes.

        Args:
            plane_points (ArrayLike):
                An array of points on planes of shape (N,3)
            plane_vectors (ArrayLike):
                An array of plane vertices of shape (N,3)
            dists (ArrayLike):
                An array of distances from the site to the planes of shape (N)
            site_cart_coords (ArrayLike):
                The cartesian coordinates of the site

        Returns:
            Three arrays corresponding to the indices of the reduced planes,
            the reduced plane's points and the reduced plane's vectors
        """
        # convert to arrays if not already
        all_plane_points = np.array(plane_points)
        all_plane_vectors = np.array(plane_vectors)
        dists = np.array(dists)
        unique_dists = list(set(dists))
        # Get initial set of planes
        initial_indices = np.where(dists == unique_dists[0])[0]
        # Create arrays of plane points and vectors
        reduced_plane_points = all_plane_points[initial_indices]
        reduced_plane_vectors = all_plane_vectors[initial_indices]
        reduced_indices = initial_indices.copy()
        # loop over every unique distance. All the planes at this distance are
        # symmetric. Check if they are important by seeing if the line between the
        # atom and the plane point is intersected by any of the existing planes
        for dist in unique_dists[1:]:
            first_index = np.where(dists == dist)[0][0]
            point0 = site_cart_coord
            point1 = all_plane_points[first_index]
            # important_plane = True
            intersects = []
            for i, plane_point in enumerate(reduced_plane_points):
                plane_vector = reduced_plane_vectors[i]
                intersect = self._get_vector_plane_intersection(
                    point0, point1, plane_point, plane_vector
                )
                if intersect is None:
                    pass
                else:
                    # important_plane = False
                    intersects.append(intersect)
                    break
            # if important_plane:
            if len(intersects) < 3:
                indices_to_add = np.where(dists == dist)[0]
                reduced_indices = np.concatenate((reduced_indices, indices_to_add))
                points_to_add = all_plane_points[indices_to_add]
                vectors_to_add = all_plane_vectors[indices_to_add]
                reduced_plane_points = np.concatenate(
                    (reduced_plane_points, points_to_add)
                )
                reduced_plane_vectors = np.concatenate(
                    (reduced_plane_vectors, vectors_to_add)
                )
        return reduced_indices, reduced_plane_points, reduced_plane_vectors

    @cached_property
    def all_site_neighbor_pairs(self):
        """
        A dataframe containing information about all site-neighbor pairs in a
        structure within 15A of each other.
        """
        grid = self.grid
        structure = grid.structure
        # logging.info("Getting all neighboring atoms for each site in structure")
        # Get all neighbors within 15 Angstrom
        nearest_neighbors = structure.get_neighbor_list(15)
        # Get the equivalent atom index for each atom
        equivalent_atoms = grid.equivalent_atoms
        equiv_site_index = equivalent_atoms[nearest_neighbors[0]]
        equiv_neigh_index = equivalent_atoms[nearest_neighbors[1]]
        # Create dataframe with important info about each site/neighbor pair
        site_neigh_pairs = pd.DataFrame()
        # Add sites and neighbors indices
        site_neigh_pairs["site_index"] = nearest_neighbors[0]
        site_neigh_pairs["neigh_index"] = nearest_neighbors[1]
        site_neigh_pairs["equiv_site_index"] = equiv_site_index
        site_neigh_pairs["equiv_neigh_index"] = equiv_neigh_index
        site_neigh_pairs["site_symbol"] = None
        site_neigh_pairs["neigh_symbol"] = None

        site_cart_coords = []
        for site_index, site in enumerate(structure):
            # Add species strings to all site and neighbor indices with given index
            species_string = site.species_string
            site_condition = site_neigh_pairs["site_index"] == site_index
            neigh_condition = site_neigh_pairs["neigh_index"] == site_index
            site_neigh_pairs.loc[site_condition, "site_symbol"] = species_string
            site_neigh_pairs.loc[neigh_condition, "neigh_symbol"] = species_string
            # Get the coordinate for this site index. Create a list of arrays made up
            # of this coordinate and add it to our site coords list
            site_coord = list(site.coords)
            atom_neigh_len = len(site_neigh_pairs.loc[site_condition])
            site_coords_array = np.tile(site_coord, (atom_neigh_len, 1))
            site_cart_coords.extend(site_coords_array)

        # Add the distances for each site-neighbor pair, then round them to 5 decimals
        site_neigh_pairs["dist"] = nearest_neighbors[3]
        site_neigh_pairs["dist"] = site_neigh_pairs["dist"].round(5)
        # Add the site coordinates
        site_neigh_pairs["site_coords"] = site_cart_coords
        # Get the fractional coordinates for each neighbor atom. Then calculate the cartesian coords
        neigh_frac_coords = (
            structure.frac_coords[nearest_neighbors[1]] + nearest_neighbors[2]
        )
        neigh_cart_coords = []
        neigh_cart_coords.extend(
            grid.get_cart_coords_from_frac_full_array(neigh_frac_coords)
        )
        # Add the neighbors cartesian coordinates
        site_neigh_pairs["neigh_coords"] = neigh_cart_coords

        # Create columns for the partitioning fraction and radius
        site_neigh_pairs["partitioning_frac"] = None
        site_neigh_pairs["radius"] = None
        site_neigh_pairs.sort_values(by="dist", inplace=True)

        return site_neigh_pairs

    def get_closest_neighbors(
        self,
        structure: Structure = None,
    ):
        """
        Function for getting the closest neighbors to an atom. Uses the CrystalNN
        class from pymatgen. This is intended to help quickly check for covalency
        in the structure and may eventually be removed.

        Args:
            structure (Structure):
                The structure to get closest neighbors for. Defaults to the
                base structure for this PartitioningToolkit instance

        Returns:
            A dictionary relating atomic sites to pymatgen neighbor objects
        """
        if structure is None:
            structure = self.grid.structure
        c = CrystalNN(search_cutoff=5)
        closest_neighbors = {}
        for i in range(len(structure)):
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore", category=UserWarning, module="pymatgen"
                )
                _, _, d = c.get_nn_data(structure, n=i)
            biggest = max(d)
            closest_neighbors[i] = d[biggest]
        return closest_neighbors

    def get_set_number_of_neighbors(self, site_index, neighbor_num: int = 26):
        """
        Function for getting the closest neighbors.

        Args:
            site_index (int):
                The atomic site of interest
            neighbor_num (int):
                The number of nearest neighbors to find

        Results:
            A dataframe relating an atoms index to its  neighbors.

        """
        # Get all possible neighbor atoms for each atom within 15 angstroms
        all_neighbors = self.all_site_neighbor_pairs.copy()
        atom_neighbors = all_neighbors.loc[
            all_neighbors["site_index"] == site_index
        ].copy()
        set_neighbors = atom_neighbors.iloc[:neighbor_num]

        return set_neighbors

    def get_site_neighbor_frac(
        self,
        site_cart_coords: ArrayLike,
        neigh_cart_coords: ArrayLike,
        site_index: int,
        neigh_index: int,
    ):
        """
        Function for getting the fraction of a line betwaeen two sites where
        the ELF is at a minimum.

        Args:
            site_cart_coords (Array | list):
                The cartesian coordinates of the first site.
            neigh_cart_coords (Array | list):
                The cartesian coordinates of the neighboring site.

        Returns:
            The minimum point in the ELF between the two sites.

        """
        grid = self.grid.copy()

        site_voxel_coord = grid.get_voxel_coords_from_cart(site_cart_coords)
        neigh_voxel_coord = grid.get_voxel_coords_from_cart(neigh_cart_coords)

        # we need a straight line between these two points.  get list of all ELF values
        (
            elf_coordinates,
            elf_values,
            label_values,
        ) = self.get_partitioning_line_from_voxels(
            site_voxel_coord, neigh_voxel_coord, method="linear"
        )

        # find the minimum position and value along the elf_line
        # the third element is the fractional position, measured from site_voxel_coord
        elf_min_index, elf_min_value, elf_min_frac = self.get_line_minimum_as_frac(
            elf_coordinates,
            elf_values,
            label_values,
            site_index,
            neigh_index,
        )
        return elf_min_frac

    @staticmethod
    def get_plane_equation(
        plane_point: ArrayLike | list,
        plane_vector: ArrayLike | list,
    ):
        """
        Gets the equation of a plane from a vector normal to the plane and a
        point on the plane

        Args:
            plane_point (ArrayLike):
                A point on the plane
            plane_vector (ArrayLike):
                A vector perpendicular to the plane

        Returns:
            The four plane equation coefficients
        """
        x0, y0, z0 = plane_point
        a, b, c = plane_vector
        d = a * x0 + b * y0 + c * z0
        return a, b, c, d

    @classmethod
    def find_intersection_point(
        cls,
        plane1: ArrayLike | list,
        plane2: ArrayLike | list,
        plane3: ArrayLike | list,
    ):
        """
        Finds the point at which three planes intersect if it exists

        Args:
            plane1 (ArrayLike):
                An Array of length 6 with the first three values representing
                the vector orthogonal to the plane and the second three values
                representing the point on the plane.
            plane2 (ArrayLike):
                An Array of length 6 with the first three values representing
                the vector orthogonal to the plane and the second three values
                representing the point on the plane.
            plane3 (ArrayLike):
                An Array of length 6 with the first three values representing
                the vector orthogonal to the plane and the second three values
                representing the point on the plane.

        Returns:
            The point at which the planes intersect as an array
        """

        a1, b1, c1, d1 = cls.get_plane_equation(plane1[:3], plane1[3:])
        a2, b2, c2, d2 = cls.get_plane_equation(plane2[:3], plane2[3:])
        a3, b3, c3, d3 = cls.get_plane_equation(plane3[:3], plane3[3:])

        A = np.array([[a1, b1, c1], [a2, b2, c2], [a3, b3, c3]])
        b = np.array([d1, d2, d3])

        # Solve the system of equations
        intersection_point = np.linalg.solve(A, b)

        return intersection_point

    @classmethod
    def get_important_planes(cls, planes: list):
        """
        Gets a list of planes that make up a 3D polyhedral.

        Args:
            planes (list):
                The list of planes to check

        Returns:
            A list of plane intercepts and a list of important planes as arrays
            with the first 3 values being the vector and the second three values
            being the point.
        """
        # create list for points where planes intersect
        intercepts = []
        important_planes = []
        # iterate through each set of 3 planes
        for combination in combinations(planes, 3):
            # try to find an intersection point. We do a try except because if two
            # of the planes are parallel there wont be an intersect point
            try:
                intercept = cls.find_intersection_point(
                    combination[0], combination[1], combination[2]
                )
            except:
                continue

            # Check if the points are one or within the convex shape defined by the
            # planes. Assume this is true at first
            important_intercept = True
            # Check each plane versus the intercept point. If we plug the point into
            # the plane equation it should return as 0 or positive if it is within the
            # shape.
            for plane in planes:
                sign, dist = cls.get_plane_sign(
                    point=intercept,
                    plane_vector=plane[3:],
                    plane_point=plane[:3],
                )
                # if sign in ["positive", "zero"]:
                # print(dist)
                if dist >= -1e-5:
                    pass
                else:
                    important_intercept = False
                    # print(dist)
                    break
            # If the point is bound by all planes, it is an important intercept.
            # append it to our list.
            if important_intercept:
                # check if intercept is already in list. If it's not, continue
                # repeat_intercept = any(
                #     np.array_equal(intercept, arr) for arr in intercepts
                # )
                # if not repeat_intercept:
                intercepts.append(intercept)
                for plane in combination:
                    repeat_plane = any(
                        np.array_equal(plane, arr) for arr in important_planes
                    )
                    # If this isn't a repeat plane, add it to our important planes list
                    if not repeat_plane:
                        important_planes.append(plane)

        return intercepts, important_planes

    @cached_property
    def maximum_site_neighbor_pairs(self):
        """
        Finds the maximum list of sites that might contribute to the partitioning
        surface. This is found by considering a small surface made up of planes
        that are definitely on the partitioning surface. A larger surface is
        then constructed by doubling the distance from the atom to each plane
        (the distance to the neighbor atom is used instead if it is larger). Then
        any site inside of this new surface may contribute to the overall surface

        Returns:
            A dataframe of site neighbor pairs with plane information
        """
        # get the structure and a list pointing to the first equivalent atom
        # for each site
        structure = self.grid.structure
        equivalent_atoms = self.grid.equivalent_atoms
        unique_atoms = list(set(equivalent_atoms))
        # get a dataframe containing all site neighbor pairs. We will reduce
        # from here
        all_site_neighbor_pairs = self.all_site_neighbor_pairs
        # Get the site neighbor pairs corresponding to each unique atom in the
        # structure. We only look at unique atoms to avoid a large number of
        # calculations in highly symmetrical structures
        unique_site_dataframes = {}
        for site_index, site in enumerate(structure):
            # check if site is unique
            if site_index in unique_atoms:
                # locate all rows in full site-neighbor pair dataframe where
                # the site matches the one we're looking at
                site_dataframe = all_site_neighbor_pairs.loc[
                    all_site_neighbor_pairs["site_index"] == site_index
                ]
                site_dataframe.reset_index(inplace=True, drop=True)
                # add the reduced dataframe to our list of unique dataframes
                unique_site_dataframes[site_index] = site_dataframe
        # Now we begin reducing the site-neighbor pairs. We will use a small
        # set of planes that enclose the site to check which other sites might
        # potentially contribute to the partitioning surface.
        # We will start with the closest neighbors, calculate the dividing
        # planes and then check if these planes enclose a reasonable space. If
        # not we'll add planes until it does. Once we have our surface planes,
        # we'll calculate which sites are inside of this surface
        logging.info("Calculating maximum set of neighboring atoms")
        atom_potential_planes = {}
        for site_index, site_dataframe in tqdm(
            unique_site_dataframes.items(),
            total=len(unique_site_dataframes),
            ascii="░▒▓",
        ):
            # First we find the unique sets of atom neighbor pairs for this atom.
            unique_site_df = site_dataframe.drop_duplicates(
                subset=["site_symbol", "neigh_symbol", "dist"]
            )
            # create our dataframe for out surface planes
            surface_planes = pd.DataFrame()
            # Now we progressively add more planes by distance from the atom unitl we
            # have a closed surface around the atom.
            for i, row in unique_site_df.iterrows():
                # get the coords of each site and its neighbor for this row
                site_cart_coords = row["site_coords"]
                neigh_cart_coords = row["neigh_coords"]
                # get the site and neighbor symbols
                neigh_index = row["neigh_index"]
                shortest_dist = row["dist"]
                # get fraction along line where the min is located
                frac = self.get_site_neighbor_frac(
                    site_cart_coords,
                    neigh_cart_coords,
                    site_index,
                    neigh_index,
                )
                # Set frac to 2*frac or 1 whichever is larger
                if 2 * frac >= 1:
                    frac = (2 * frac) + 0.1
                else:
                    frac = 1.1
                # get all of the planes at this distance
                planes_at_dist = site_dataframe.loc[
                    site_dataframe["dist"] == shortest_dist
                ].copy()

                # Get the site and neighbor coords in arrays to make calculations easier
                site_coords = np.array(planes_at_dist["site_coords"].to_list())
                neigh_coords = np.array(planes_at_dist["neigh_coords"].to_list())

                # Calculate the plane points and vectors for this set of planes
                vectors = neigh_coords - site_coords
                magnitudes = np.linalg.norm(vectors, axis=1)
                unit_vectors = vectors / magnitudes[:, np.newaxis]
                plane_points = vectors * frac + site_coords
                # Append them to the initial vectors dataframe
                planes_at_dist["frac"] = frac
                planes_at_dist["plane_vectors"] = list(unit_vectors)
                planes_at_dist["plane_points"] = list(plane_points)
                surface_planes = pd.concat([surface_planes, planes_at_dist])
                # Get all of the surface planes so far
                surface_plane_points = np.array(
                    surface_planes["plane_points"].to_list()
                )
                surface_plane_vectors = np.array(
                    surface_planes["plane_vectors"].to_list()
                )
                surface_planes_array = np.concatenate(
                    (surface_plane_points, surface_plane_vectors), axis=1
                )
                # Calculate the points at which the three planes intersect to
                # form a closed polyhedra
                important_vertices, _ = self.get_important_planes(surface_planes_array)
                try:
                    # check if we've enclosed a reasonably sized space. If not
                    # repeat with more planes. Otherwise continue on
                    hull = ConvexHull(important_vertices)
                    volume = hull.volume
                    if volume < 1200:
                        break
                except:
                    continue
            # Now we need to find which atoms are contained in this bound region. We loop
            # over each unique atom-neigh pair. If the neighbor is within the region, it
            # is possible this neighbor and all symmetrically equivalent ones are important
            # to the voronoi surface
            potential_plane_indices = []
            for i, row in unique_site_df.iterrows():
                neigh_cart_coords = row["neigh_coords"]
                dist = row["dist"]
                potential_bounding_plane = True
                # Check each plane versus the intercept point. If we plug the point into
                # the plane equation it should return as 0 or positive if it is within the
                # shape?
                for plane in surface_planes_array:
                    sign, _ = PartitioningToolkit.get_plane_sign(
                        point=neigh_cart_coords,
                        plane_vector=plane[3:],
                        plane_point=plane[:3],
                    )
                    if sign in ["positive", "zero"]:
                        pass
                    else:
                        potential_bounding_plane = False
                        break
                if potential_bounding_plane:
                    # These are potentially important planes. We want to store the indices
                    # So they're easy to find for every identical atom
                    indices = site_dataframe.loc[
                        site_dataframe["dist"] == dist
                    ].index.to_list()
                    potential_plane_indices.extend(indices)

            atom_potential_planes[site_index] = potential_plane_indices

        # Now we get all possible site neighbor pairs in one dataframe, similar
        # to the all_site_neighbor_pairs dataframe but reduced
        possible_site_neigh_pairs = pd.DataFrame()
        for site_index, site in enumerate(structure):
            equivalent_atom = self.grid.equivalent_atoms[site_index]
            plane_indices = atom_potential_planes[equivalent_atom]
            site_dataframe = all_site_neighbor_pairs.loc[
                all_site_neighbor_pairs["site_index"] == site_index
            ].copy()
            site_dataframe.sort_values(by="dist", inplace=True)
            site_dataframe.reset_index(inplace=True, drop=True)
            reduced_site_dataframe = site_dataframe.iloc[plane_indices]
            possible_site_neigh_pairs = pd.concat(
                [possible_site_neigh_pairs, reduced_site_dataframe]
            )
        possible_site_neigh_pairs = self._calculate_partitioning_planes(
            possible_site_neigh_pairs
        )
        return possible_site_neigh_pairs

    def _calculate_partitioning_planes(self, possible_site_neigh_pairs):
        """
        Calculates the plane points and vectors for provided site neighbor pairs.

        Args:
            possible_site_neigh_pairs (pd.DataFrame):
                A dataframe of potential site neighbor pairs. Must have the
                columns "site_coords, "neigh_coords", "site_symbol",
                "neigh_symbol", and "dist".

        Returns:
            A dataframe of site neighbor pairs with plane information
        """
        possible_unique_pairs = possible_site_neigh_pairs.drop_duplicates(
            # subset=["site_symbol", "neigh_symbol", "dist"]
            subset=["equiv_site_index", "equiv_neigh_index", "dist"]
        )

        # calculate fractions along each line in the ELF
        logging.info("Calculating partitioning plane positions from ELF")
        # Get partitioning frac for each unique site_neighbor pair
        for index, row in tqdm(
            possible_unique_pairs.iterrows(),
            total=len(possible_unique_pairs),
            ascii="░▒▓",
        ):
            # Check if we've already found the frac for this row
            if row["partitioning_frac"] is None:
                # get coords of each site and its neighbor
                site_cart_coords = row["site_coords"]
                # site_voxel_coords = grid.get_voxel_coords_from_cart(site_cart_coords)
                neigh_cart_coords = row["neigh_coords"]
                # neigh_voxel_coords = grid.get_voxel_coords_from_cart(neigh_cart_coords)
                # Get the site symbols.
                # needed to update the dataframe
                site_index = row["site_index"]
                neigh_index = row["neigh_index"]
                dist = row["dist"]

                # get fraction along line where the min is located
                frac = self.get_site_neighbor_frac(
                    site_cart_coords,
                    neigh_cart_coords,
                    site_index,
                    neigh_index,
                )
                radius = frac * dist
                reverse_frac = 1 - frac
                reverse_radius = reverse_frac * dist

                # Get the unique indices to search the dataframe with
                equiv_site_index = row["equiv_site_index"]
                equiv_neigh_index = row["equiv_neigh_index"]

                # create search to find rows with same symbol set and reverse symbol set.
                reverse_condition = (
                    (possible_unique_pairs["equiv_site_index"] == equiv_neigh_index)
                    & (possible_unique_pairs["equiv_neigh_index"] == equiv_site_index)
                    & (possible_unique_pairs["dist"] == dist)
                )
                # assign the fraction along the line and distance to each unique
                # site neighbor pair. We do this in the loop so that the reverse
                # assignments don't need to be repeated
                possible_unique_pairs.at[index, "partitioning_frac"] = frac
                possible_unique_pairs.loc[reverse_condition, "partitioning_frac"] = (
                    reverse_frac
                )

                # create another search condition for the full dataframe of site-neighbor pairs
                search_condition1 = (
                    (possible_site_neigh_pairs["equiv_site_index"] == equiv_site_index)
                    & (
                        possible_site_neigh_pairs["equiv_neigh_index"]
                        == equiv_neigh_index
                    )
                    & (possible_site_neigh_pairs["dist"] == dist)
                )
                reverse_condition1 = (
                    (possible_site_neigh_pairs["equiv_site_index"] == equiv_neigh_index)
                    & (
                        possible_site_neigh_pairs["equiv_neigh_index"]
                        == equiv_site_index
                    )
                    & (possible_site_neigh_pairs["dist"] == dist)
                )

                # Assign the fraction along the line and distance to every
                # site neighbor pair.
                possible_site_neigh_pairs.loc[
                    search_condition1, ["partitioning_frac", "radius"]
                ] = (frac, radius)
                possible_site_neigh_pairs.loc[
                    reverse_condition1, ["partitioning_frac", "radius"]
                ] = (reverse_frac, reverse_radius)
        # Get the site and neighbor coords in arrays to make calculations easier
        site_coords = np.array(possible_site_neigh_pairs["site_coords"].to_list())
        neigh_coords = np.array(possible_site_neigh_pairs["neigh_coords"].to_list())
        fracs = possible_site_neigh_pairs["partitioning_frac"].to_numpy()

        # Calculate the plane points and vectors
        vectors = neigh_coords - site_coords
        magnitudes = np.linalg.norm(vectors, axis=1)
        unit_vectors = vectors / magnitudes[:, np.newaxis]
        plane_points = vectors * fracs[:, np.newaxis] + site_coords

        # Add plane points and vectors to full dataframe
        possible_site_neigh_pairs["plane_points"] = list(plane_points)
        possible_site_neigh_pairs["plane_vectors"] = list(unit_vectors)
        return possible_site_neigh_pairs

    def reduce_to_symmetric_partitioning(self, initial_partitioning: dict):
        """
        Reduces a set of partitioning planes by checking that each plane has
        an equivalent counterpart for the neighboring atom and removing it
        if not.

        Args:
            initial_partitioning (dict):
                A dictionary with site indices as keys and partitioning dataframes
                as values

        Returns:
            A new dictionary of partitioning dataframes for each site.
        """
        equivalent_atoms = self.grid.equivalent_atoms
        unique_atoms = list(set(equivalent_atoms))

        planes_to_keep = {}

        for atom in unique_atoms:
            partitioning_df = initial_partitioning[atom]
            atom_planes_to_keep = []
            # check each plane in the partitioning df
            for i, row in partitioning_df.iterrows():
                neigh_index = row["neigh_index"]
                # get the partitioning to check against
                equiv_neigh_index = equivalent_atoms[neigh_index]
                # get the distance for this site neighbor pair
                dist = row["dist"]
                neigh_partitioning_df = initial_partitioning[equiv_neigh_index]
                # if the reverse exists in the neighbors partitioning keep this
                # plane.
                if dist in neigh_partitioning_df["dist"].to_list():
                    atom_planes_to_keep.append(i)
            planes_to_keep[atom] = atom_planes_to_keep

        new_partitioning = {}
        for i, partitioning_df in initial_partitioning.items():
            equivalent_atom = equivalent_atoms[i]
            indices = planes_to_keep[equivalent_atom]
            new_partitioning_df = partitioning_df.iloc[indices].copy()
            new_partitioning_df.sort_values(by=["dist"], inplace=True)
            new_partitioning_df.reset_index(inplace=True, drop=True)
            new_partitioning[i] = new_partitioning_df
        return new_partitioning

    def increase_to_symmetric_partitioning(self, initial_partitioning: dict):
        """
        Adds to a set of partitioning planes by checking that each plane has
        an equivalent counterpart for the neighboring atom and adding it
        if not.

        Args:
            initial_partitioning (dict):
                A dictionary with site indices as keys and partitioning dataframes
                as values

        Returns:
            A new dictionary of partitioning dataframes for each site.
        """
        equivalent_atoms = self.grid.equivalent_atoms
        site_indices = [i for i in range(len(self.grid.structure))]
        new_partitioning = {}
        for site in site_indices:
            # Look through each other unique atom.
            part_df = initial_partitioning[site]
            site_coords = part_df.loc[0, "site_coords"]
            for neigh in site_indices:
                if neigh != site:
                    neigh_part_df = initial_partitioning[neigh]
                    # Get the planes connecting this neighbor to the site
                    important_neigh_df = neigh_part_df.loc[
                        neigh_part_df["neigh_index"] == site
                    ]
                    # For each plane we want to do the following:
                    # 1. Check if the original site has this plane
                    # 2. Transform the coords of the neigh
                    # 3. Flip the partitioning frac and radius
                    # 4. transform plane point and flip plane vector
                    for i, row in important_neigh_df.iterrows():
                        dist = row["dist"]
                        condition = (part_df["dist"] == dist) & (
                            part_df["neigh_index"] == neigh
                        )
                        if len(part_df.loc[condition]) == 0:
                            # We want to add a new row
                            transformation_vector = row["neigh_coords"] - site_coords
                            new_row = {
                                "site_index": site,
                                "neigh_index": neigh,
                                "equiv_site_index": equivalent_atoms[site],
                                "equiv_neigh_index": equivalent_atoms[neigh],
                                "site_symbol": row["neigh_symbol"],
                                "neigh_symbol": row["site_symbol"],
                                "dist": dist,
                                "site_coords": site_coords,
                                "neigh_coords": row["site_coords"]
                                + transformation_vector,
                                "partitioning_frac": 1 - row["partitioning_frac"],
                                "radius": dist - row["radius"],
                                "plane_points": row["plane_points"]
                                + transformation_vector,
                                "plane_vectors": row["plane_vectors"] * -1,
                            }
                            part_df.loc[len(part_df)] = new_row
            part_df.sort_values(by=["dist"], inplace=True)
            part_df.reset_index(inplace=True, drop=True)
            new_partitioning[site] = part_df
        return new_partitioning

    def get_partitioning(self):
        """
        Gets the partitioning planes for each atom as well as some other useful
        information.

        Returns:
            A dictionary where the keys are site indices and the values
            are neighbor dictionaries containing information on the partitioning
            planes.
        """
        structure = self.grid.structure
        equivalent_atoms = self.grid.equivalent_atoms
        unique_atoms = list(set(equivalent_atoms))
        possible_site_neigh_pairs = self.maximum_site_neighbor_pairs
        # seperate the possible site neighbor pairs into individual atomic sites
        site_dataframes = []
        for site_index, site in enumerate(structure):
            site_dataframe = possible_site_neigh_pairs.loc[
                possible_site_neigh_pairs["site_index"] == site_index
            ].copy()
            site_dataframe.sort_values(by="dist", inplace=True)
            site_dataframes.append(site_dataframe)

        # Now we want to reduce the planes more rigorously by finding where each
        # set of three planes intersect and determining if these points are
        # important to the partitioning surface. We only do this for each unique
        # atom and then we'll fill in the other atoms afterwards
        logging.info("Reducing to necessary partitioning planes")
        important_plane_indices = {}
        for site_index in tqdm(unique_atoms, total=len(unique_atoms), ascii="░▒▓"):
            # get the plane dataframe for this site
            site_dataframe = site_dataframes[site_index]
            # get the plane points and vectors for this site
            plane_points = site_dataframe["plane_points"].to_list()
            plane_vectors = site_dataframe["plane_vectors"].to_list()
            planes = np.concatenate(
                (np.array(plane_points), np.array(plane_vectors)), axis=1
            )
            planes[np.where(planes == 0)] = 1e-12
            dists = site_dataframe["dist"]
            site_cart_coord = structure[site_index].coords
            # reduce the planes by checking if any of the line segments
            # between the planes and sites are intersected by more than 2 planes
            # other than the corresponding plane
            _, reduced_plane_points, reduced_plane_vectors = self._reduce_planes_quick(
                plane_points, plane_vectors, dists, site_cart_coord
            )
            reduced_planes = np.concatenate(
                (reduced_plane_points, reduced_plane_vectors), axis=1
            )
            reduced_planes[np.where(reduced_planes == 0)] = 1e-12
            important_planes = reduced_planes.copy()
            # Get the important planes more rigorously by checking which planes
            # contribute to the vertices of the polyhedral shape surrounding
            # the site
            # _, reduced_planes = self.get_important_planes(reduced_planes)
            # reduced_planes = np.array(reduced_planes)
            # get the indices for the important planes and append them to our
            # dictionary
            indices = np.where((planes[:, None] == important_planes).all(axis=2))[1]
            important_plane_indices[site_index] = indices

        # Go through each site, get the corresponding important planes using
        # the indices we just found, and add to a partitioning dataframe
        initial_partitioning = {}
        for site_index, site_dataframe in enumerate(site_dataframes):
            equiv_atom = equivalent_atoms[site_index]
            important_plane_index = important_plane_indices[equiv_atom]
            site_dataframe = site_dataframe.iloc[important_plane_index].copy()
            site_dataframe.sort_values(by=["dist"], inplace=True)
            site_dataframe.reset_index(inplace=True, drop=True)
            initial_partitioning[site_index] = site_dataframe

        # partitioning = self.reduce_to_symmetric_partitioning(
        #     initial_partitioning
        # )
        # partitioning = self.increase_to_symmetric_partitioning(initial_partitioning)
        # return partitioning
        return initial_partitioning

    def plot_partitioning_results(
        self,
        partition_results: dict = None,
    ):
        """
        Plots the unit cell and partitioning planes from the partitioning
        results.

        Args:
            partition_results (dict): The results from running the partitioning
                algorithm
        """
        if partition_results is None:
            partition_results, _ = self.get_partitioning()

        # Create a matplotlib plot
        import matplotlib
        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(projection="3d")

        # Get the vertices of the polygons formed by the partitioning around
        # each atom
        atoms_polygon_nodes = {}
        for site_index, neighbor_df in partition_results.items():
            planes = []
            for neigh_index, row in neighbor_df.iterrows():
                plane_point = list(row["plane_points"])
                plane_vector = list(row["plane_vectors"])
                plane = plane_vector + plane_point
                planes.append(plane)
            intercepts, _ = self.get_important_planes(planes)
            atoms_polygon_nodes[site_index] = intercepts

        # get the structure and species
        structure = self.grid.structure
        species = structure.symbol_set

        # plot the unit cell
        self.grid._plot_unit_cell(ax, fig)

        # get a color map to match all same atoms to the same color
        color_map = matplotlib.colormaps.get_cmap("tab10")

        # go through each atom species, set the unique color
        for i, specie in enumerate(species):
            color = color_map(i)
            site_indices = structure.indices_from_symbol(specie)
            # for each site, loop through and plot the surfaces.
            for site in site_indices:
                hull = ConvexHull(atoms_polygon_nodes[site])
                triangles = hull.simplices

                x = []
                y = []
                z = []
                for point in atoms_polygon_nodes[site]:
                    x.append(point[0])
                    y.append(point[1])
                    z.append(point[2])

                ax.plot_trisurf(x, y, z, triangles=triangles, color=color)



# Generated by Django 4.2.7 on 2024-02-26 19:58

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("badelf", "0002_badelf_electrides_per_formula"),
    ]

    operations = [
        migrations.AddField(
            model_name="badelf",
            name="electrides_per_reduced_formula",
            field=models.FloatField(blank=True, null=True),
        ),
    ]
